"""
OpenStreetMap importer using OSMnx and neatnet.
Work in progress. Import from OSM is experimental and may not work as expected. It is functional but may produce inappropriate networks for simulation, such as too many nodes, too many deadends, fragmented networks.

Confirmed version: neatnet 0.1.0
"""

import sys
import warnings
from collections import defaultdict, Counter
import numpy as np
import pandas as pd
import networkx as nx


if 1: #to avoid editor's failure
    if sys.version_info < (3, 11):
        raise RuntimeError("Python 3.11 or higher is required to use 'OSMImporter2'.")

try:
    import osmnx as ox
except:
    raise ImportError("Optional module 'osmnx' is not installed. Use 'pip install uxsim[advanced]'.")

try:
    import neatnet # type: ignore
except:
    raise ImportError("Optional module 'neatnet' is not installed. Use 'pip install uxsim[advanced]'. Note that 'neatnet' requires Python 3.11 or later.")

try:
    import geopandas as gpd
except:
    raise ImportError("Optional module 'geopandas' is not installed. Use 'pip install uxsim[advanced]'.")

try:
    from shapely.geometry import Point, LineString
except:
    raise ImportError("Optional module 'shapely' is not installed. Use 'pip install uxsim[advanced]'.")


def _obtain_simplify_osm_by_osmnx_neatnet(bbox, custom_filter, simplification, kwargs_dict_for_osmnx_graph_from_bbox, kwargs_dict_for_osmnx_project_graph, kwargs_dict_for_neatnet_neatify):
    """
    Obtains and simplifies OSM data using OSMnx and neatnet libraries. This function retrieves a road network from OpenStreetMap within the specified bounding box, applies projection, and then simplifies the network links using the neatnet library's neatify algorithm.

    Parameters
    ----------
    bbox : tuple
        Bounding box coordinates expected by OSMnx.
    custom_filter : str
        OSM filter query to use when downloading the network data.

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing the simplified road network links.

    Notes
    -----
    This function requires the OSMnx and neatnet libraries to be installed.
    The returned network is filtered for driving ("drive" network type).
    """
    
    G = ox.graph.graph_from_bbox(bbox=bbox, network_type="drive", custom_filter=custom_filter, **kwargs_dict_for_osmnx_graph_from_bbox)
    G_projected = ox.project_graph(G, **kwargs_dict_for_osmnx_project_graph)

    if simplification:
        simplified_links = neatnet.neatify(ox.graph_to_gdfs(G_projected)[1], **kwargs_dict_for_neatnet_neatify)
    else:
        simplified_links = ox.graph_to_gdfs(G_projected)[1]
    
    return simplified_links


def _extract_network_from_geodataframe(gdf):
    """
    Function to extract road network nodes and links from a GeoDataFrame.
    
    Parameters:
    -----------
    gdf : GeoDataFrame
        GeoDataFrame containing linear features (LineString)
    
    Returns:
    --------
    nodes_df : GeoDataFrame
        Node list (node_id, x, y, geometry)
    links_df : GeoDataFrame
        Link list (link_id, from_node, to_node, length, and all other columns from the original GeoDataFrame)
    """
    # LineStringの形状のみを処理
    if not all(isinstance(geom, LineString) for geom in gdf.geometry):
        raise ValueError("GeoDataFrameにはLineString型のジオメトリのみが含まれている必要があります")
    
    # ノードの収集（始点と終点）
    nodes = []
    node_id_map = {}  # 座標からノードIDへのマッピング
    node_counter = 0
    
    # リンク情報の収集
    links = []
    
    for idx, row in gdf.iterrows():
        line = row.geometry
        
        # 始点の座標を取得
        start_point = line.coords[0]
        # 終点の座標を取得
        end_point = line.coords[-1]
        
        # 始点が既に登録されているかチェック
        start_key = (round(start_point[0], 6), round(start_point[1], 6))
        if start_key not in node_id_map:
            node_id_map[start_key] = node_counter
            nodes.append({
                'node_id': node_counter,
                'x': start_point[0],
                'y': start_point[1],
                'geometry': Point(start_point)
            })
            node_counter += 1
        
        # 終点が既に登録されているかチェック
        end_key = (round(end_point[0], 6), round(end_point[1], 6))
        if end_key not in node_id_map:
            node_id_map[end_key] = node_counter
            nodes.append({
                'node_id': node_counter,
                'x': end_point[0],
                'y': end_point[1],
                'geometry': Point(end_point)
            })
            node_counter += 1
        
        # リンク情報の作成（元のGeoDataFrameの全カラムを保持）
        link_data = {
            'link_id': idx,
            'from_node': node_id_map[start_key],
            'to_node': node_id_map[end_key],
            'length': line.length,
        }
        
        # 元のGeoDataFrameの全カラムをコピー
        for col in gdf.columns:
            if col not in ['geometry', 'link_id', 'from_node', 'to_node', 'length']:  # ジオメトリ等は別途処理するので除外
                link_data[col] = row[col]
                
        # ジオメトリはリンクデータに追加
        link_data['geometry'] = line
        
        links.append(link_data)
    
    # ノードとリンクのDataFrameを作成
    nodes_df = pd.DataFrame(nodes)
    links_df = pd.DataFrame(links)
    
    # GeoDataFrameに変換
    nodes_gdf = gpd.GeoDataFrame(nodes_df, geometry='geometry', crs=gdf.crs)
    links_gdf = gpd.GeoDataFrame(links_df, geometry='geometry', crs=gdf.crs)
    
    return nodes_gdf, links_gdf


def _attach_osm_nodes_links(W, nodes, links, set_node_capacity, override_osm_attributes=False, maxspeed_by_road_type={}, lanes_by_road_type={}):
    """
    Adds OSM nodes and links to the network. This function processes the OSM nodes and links dataframes and adds them to the provided network object (W). For nodes, it extracts coordinates and ID. For links, it processes attributes like speed limits, number of lanes, and handles one-way/two-way streets accordingly.

    Parameters
    ----------
    W : Network object
        The network object to which nodes and links will be added.
    nodes : pandas.DataFrame
        DataFrame containing node information with columns:

        - node_id: unique identifier for each node
        - x: x-coordinate
        - y: y-coordinate

    links : pandas.DataFrame
        DataFrame containing link information with columns:

        - from_node: source node ID
        - to_node: destination node ID
        - length: length of the link
        - maxspeed: maximum speed (km/h)
        - lanes: number of lanes
        - name: name of the link
        - oneway: boolean indicating if the link is one-way
    """

    maxspeed_dict = {
        "motorway": 100,
        "trunk": 80,
        "primary": 60,
        "secondary": 50,
        "tertiary": 40,
        "residential": 30,
        "living_street": 20,
        "service": 20,
        "others": 40
    }
    maxspeed_dict = {key:maxspeed_dict[key]/3.6 for key in maxspeed_dict.keys()}
    for key in maxspeed_by_road_type:
        maxspeed_dict[key] = maxspeed_by_road_type[key]

    lanes_dict = {
        "motorway": 3,
        "trunk": 3,
        "primary": 2,
        "secondary": 2,
        "tertiary": 2,
        "residential": 1,
        "living_street": 1,
        "service": 1,
        "others": 1
    }
    for key in lanes_by_road_type:
        lanes_dict[key] = lanes_by_road_type[key]

    added_node_names = []
    for i in nodes.index:
        x = nodes.at[i, "x"]
        y = nodes.at[i, "y"]
        node_id = str(nodes.at[i, "node_id"])
        W.addNode(node_id, x, y, auto_rename=True)

        added_node_names.append(node_id)

    for i in links.index:
        start_node = str(links.at[i, "from_node"])
        end_node = str(links.at[i, "to_node"])
        length = links.at[i, "length"]

        name = links.at[i, "name"]
        try:
            if type(name) is list:
                name = name[0]
            if type(name) is not str:
                name = str(i)
        except:
            name = str(i)
        
        highway_type = links.at[i, "highway"]
        try:
            if highway_type == None:
                #estimate the type from connected links
                from_node = links.at[i, "from_node"]
                links_upstream = links[links["to_node"] == from_node]
                to_node = links.at[i, "from_node"]
                links_downstream = links[links["from_node"] == to_node]
                
                links_connected = list(links_upstream["highway"])+list(links_downstream["highway"])
                links_connected_filtered = [l for l in links_connected if l != None]

                if len(links_connected_filtered):
                    assumed_type = Counter(links_connected_filtered).most_common(1)[0][0]
                else:
                    assumed_type = "None"
                
                links.at[i, "highway"] = assumed_type
                highway_type = str(assumed_type)
                
            if type(highway_type) is list:
                highway_type = str(Counter(highway_type).most_common(1)[0][0])
                links.at[i, "highway"] = highway_type
                
        except Exception as e:
            highway_type = "None"
            warnings.warn(f"Link {name} - Could not parse highway_type: {links.at[i, 'highway']}, due to '{e}' error. Using default value of {'None'}", UserWarning)

        maxspeed = links.at[i, "maxspeed"]
        try:
            if type(maxspeed) is str and not override_osm_attributes:
                maxspeed = float(maxspeed)/3.6
            elif type(maxspeed) is list and not override_osm_attributes:
                maxspeed = np.average([float(v) for v in maxspeed])/3.6
            else:
                maxspeed = maxspeed_dict["others"]
                for key in maxspeed_dict:
                    if key in highway_type:
                        maxspeed = maxspeed_dict[key]
                        break
        except Exception as e:
            maxspeed = maxspeed_dict["others"]
            warnings.warn(f"Link {name} - Could not parse maxspeed: {links.at[i, 'maxspeed']}, due to '{e}' error. Using default value of {maxspeed_dict['others']} m/s", UserWarning)
        

        """
        東京の一般道で確認できた範囲では，oneway=Falseの時のlanesは両方向の総和で，oneway=Trueの時のlanesはその方向のみの数
        """
        oneway = links.at[i, "oneway"]
        lanes = links.at[i, "lanes"]
        try:
            if type(lanes) is str and not override_osm_attributes:
                lanes = int(lanes)
            elif type(lanes) is list and not override_osm_attributes:
                lanes = max([int(l) for l in lanes])
            else:
                lanes = lanes_dict["others"]
                for key in lanes_dict:
                    if key in highway_type:
                        if oneway:
                            lanes = lanes_dict[key]
                        else:
                            lanes = lanes_dict[key]*2
                        break
        except Exception as e:
            lanes = lanes_dict["others"]
            warnings.warn(f"Link {name} - Could not parse number of lane: {links.at[i, 'lanes']}, due to '{e}' error. Using default value of {lanes_dict['others']}", UserWarning)

        if oneway:
            W.addLink(name=name, start_node=start_node, end_node=end_node, length=length, free_flow_speed=maxspeed, number_of_lanes=lanes, auto_rename=True)
        else:
            lanes = int(lanes/2)
            if lanes <= 0:
                lanes = 1
            W.addLink(name=name, start_node=start_node, end_node=end_node, length=length, free_flow_speed=maxspeed, number_of_lanes=lanes, auto_rename=True)
            W.addLink(name=name+"-reverse", start_node=end_node, end_node=start_node, length=length, free_flow_speed=maxspeed, number_of_lanes=lanes, auto_rename=True)
    
    if set_node_capacity:
        for node in W.NODES:
            if node.name in added_node_names:
                max_lanes = max([l.number_of_lanes for l in node.inlinks.values()] + [l.number_of_lanes for l in node.outlinks.values()])
                node.flow_capacity = max_lanes*0.8*2
                node.number_of_lanes = max_lanes*2


def osm_network_to_World(W, north=None, south=None, east=None, west=None, bbox=None, custom_filter='["highway"~"trunk|primary"]', 
                         simplification=True, override_osm_attributes=False,
                         set_node_capacity=True, 
                         maxspeed_by_road_type={}, lanes_by_road_type={}, kwargs_dict_for_osmnx_graph_from_bbox={}, kwargs_dict_for_osmnx_project_graph={}, kwargs_dict_for_neatnet_neatify={}):
    """
    Imports OpenStreetMap (OSM) network data into a World object. This function retrieves OSM road network data within the specified bounding box, simplifies the network, and extracts nodes and links using OSMnx and neatnet. Then it attaches them to the World object.

    Parameters
    ----------
    W : World
        The World object to which the OSM network will be added.
    north, south, east, west : float, optional
        The latitudes and longitudes defining the boundaries of the area to be imported.
    bbox : list, optional
        The bounding box of the area to be imported. Format is [west, south, east, north].
        This parameter takes precedence over individual north, south, east, west arguments.
    custom_filter : str, default='["highway"~"trunk|primary"]'
        The OSM filter to be used for importing the data.
        Examples:

        - '["highway"~"motorway"]': highways only
        - '["highway"~"motorway|trunk|primary|secondary|tertiary"]': highways and major/mid arterials

    simplification : bool, default=True
        If True, the network is simplified using neatnet's neatify function.
        If False, the original network is used.
    override_osm_attributes : bool, default=False
        If True, the function will override OSM attributes (maxspeed, lanes) with default values. This can be useful when OSM data is erroneous or inconsistent.
        If False, it will use the attributes from the OSM data.
    set_node_capacity : bool, default=True
        If True, automatically determines flow capacity of nodes based on connected links.
        If False, the capacity is unbounded. This simulates traffic signals in a continuous manner.
    maxspeed_by_road_type : dict, default={}
        Dictionary to override default speed limits for different road types.
    lanes_by_road_type : dict, default={}
        Dictionary to override default number of lanes for different road types.
    kwargs_dict_for_osmnx_graph_from_bbox : dict, default={}
        Additional keyword arguments to pass to OSMnx's graph.graph_from_bbox function.
    kwargs_dict_for_osmnx_project_graph : dict, default={}
        Additional keyword arguments to pass to OSMnx's project_graph function.
    kwargs_dict_for_neatnet_neatify : dict, default={}
        Additional keyword arguments to pass to neatnet's neatify function.

    Returns
    -------
    None
        The function modifies the World object in-place by adding OSM nodes and links.
    
    Notes
    -----
    The function performs three main steps:

    1. Obtains and simplifies the OSM network using OSMnx and neatnet
    2. Extracts node and link information from the simplified network
    3. Attaches the extracted nodes and links to the World object

    """
    print("Started network import from OSM. It may take some time...")
    
    warnings.warn("Import from OSM is experimental and may not work as expected. It is functional but produces inappropriate networks for simulation, such as too many nodes, too many deadends, fragmented networks.")
    
    if bbox == None:
        bbox = [west,south,east,north]
        
    simplified_links = _obtain_simplify_osm_by_osmnx_neatnet(bbox, custom_filter, simplification,
                                                             kwargs_dict_for_osmnx_graph_from_bbox, kwargs_dict_for_osmnx_project_graph, kwargs_dict_for_neatnet_neatify)
    nodes, links = _extract_network_from_geodataframe(simplified_links)  
    _attach_osm_nodes_links(W, nodes, links, set_node_capacity, override_osm_attributes=override_osm_attributes, maxspeed_by_road_type=maxspeed_by_road_type, lanes_by_road_type=lanes_by_road_type)

    stats_count = defaultdict(lambda: 0)
    stats_length = defaultdict(lambda: 0)
    for i in links.index:
        length = links.at[i, "length"]
        highway_type = str(links.at[i, "highway"]).replace("_link", "")
        if links.at[i, "oneway"]:
            coef = 1
        else:
            coef = 2
        stats_count[highway_type] += 1*coef
        stats_length[highway_type] += length*coef

    print(f"Imported {len(nodes)} nodes and {sum([val for val in stats_count.values()])} links from OSM")
    for key in stats_count:
        print(f" '{key}' links:\t {stats_count[key]} with total length {stats_length[key]:.0f} m")




"""
This script tests various other functions in UXsim.
"""

import pytest
from numpy import *
from uxsim import *
from uxsim.Utilities import *
import dill as pickle



def test_analyzer():
    import matplotlib
    matplotlib.use('Agg')
    # Define the main simulation
    # Units are standardized to seconds (s) and meters (m)
    W = World(
        name="",    # Scenario name. Can be blank. Used as the folder name for saving results.
        deltan=5,   # Simulation aggregation unit Δn. Defines how many vehicles are grouped together (i.e., platoon size) for computation. Computation cost is generally inversely proportional to deltan^2.
        tmax=1200,  # Total simulation time (s)
        print_mode=1, save_mode=1, show_mode=0,    # Various options. print_mode determines whether to print information. Usually set to 1, but recommended 0 when running multiple simulations automatically. save_mode determines if visualization results are saved. show_mode determines if visualization results are displayed. It's good to set show_mode=1 on Jupyter Notebook, otherwise recommended 0.
        random_seed=0    # Set the random seed. Specify if you want repeatable experiments. If not, set to None. On Jupyter Notebook, randomness might not always be consistent (requires a fix).
    )

    # Define the scenario
    # Merge network: Example of hard-coded definition
    W.addNode("orig1", 0, 0) # Create a node. Parameters: node name, visualization x-coordinate, visualization y-coordinate
    node_orig2 = W.addNode("orig2", 0, 2) # W.addNode returns the created node instance, so it can be assigned to a variable
    W.addNode("merge", 1, 1)
    W.addNode("dest", 2, 1)
    W.addLink("link1", "orig1", "merge", length=1000, free_flow_speed=20, jam_density=0.2, merge_priority=0.5) # Create a link. Parameters: link name, start node, end node, length, free_flow_speed, jam_density, merge_priority during merging
    W.addLink("link2", node_orig2, "merge", length=1000, free_flow_speed=20, jam_density=0.2, merge_priority=2) # Nodes can be specified using the variable name instead of the string name
    link3 = W.addLink("link3", "merge", "dest", length=1000, free_flow_speed=20, jam_density=0.2) # W.addLink also returns the created link instance
    W.adddemand("orig1", "dest", 0, 1000, 0.4) # Create OD traffic demand. Parameters: origin node, destination node, start time, end time, demand flow rate
    W.adddemand("orig2", "dest", 500, 1000, 0.6)

    # Execute the simulation
    # Run the simulation to the end
    W.exec_simulation()

    # Run the simulation for a specific time (if you want to intervene during simulation)
    # while W.check_simulation_ongoing():
    #    W.exec_simulation(duration_t=100) # Run the simulation in 100-second chunks

    # Visualization of results: Some methods are very slow. Not necessary for the simulation functionality, so can be omitted.
    W.analyzer.print_simple_stats()
    W.analyzer.time_space_diagram_density()
    W.analyzer.time_space_diagram_traj()
    W.analyzer.time_space_diagram_traj_links([["link1", "link3"], ["link2", "link3"]])
    W.analyzer.cumulative_curves()
    W.analyzer.plot_vehicle_log("110")
    W.analyzer.plot_vehicles_log(["100", "110"])
    W.analyzer.macroscopic_fundamental_diagram()
    W.analyzer.network_average()
    for t in list(range(0,W.TMAX,int(W.TMAX/3))):
        W.analyzer.network(t, detailed=0, network_font_size=0, figsize=(4,4))
        W.analyzer.network(t, detailed=0, network_font_size=0, figsize=(4,4), state_variables="flow_speed", legend=False)
    for t in list(range(0,W.TMAX,int(W.TMAX/3))):
        W.analyzer.network(t, detailed=1, network_font_size=0)
        W.analyzer.network(t, detailed=1, network_font_size=0, state_variables="flow_speed", legend=True)
    W.analyzer.network_anim(animation_speed_inverse=15, detailed=0, network_font_size=0, legend=False)
    W.analyzer.network_anim(detailed=1, network_font_size=0, figsize=(12,12))
    W.analyzer.network_anim(detailed=0, network_font_size=0, figsize=(12,12), state_variables="flow_speed")
    W.analyzer.network_fancy(animation_speed_inverse=15, sample_ratio=0.3, interval=5, trace_length=5)

    W.analyzer.network_pillow(600, state_variables="density_speed")
    W.analyzer.network_pillow(600, state_variables="density_flow")

    # Convert results to pandas.DataFrame for easier analysis
    print(W.analyzer.basic_to_pandas()) # Basic statistics
    print(W.analyzer.od_to_pandas())    # Information by OD
    print(W.analyzer.mfd_to_pandas())   # MFD (Macroscopic Fundamental Diagram)
    print(W.analyzer.link_to_pandas())  # Information per link
    print(W.analyzer.link_traffic_state_to_pandas())    # Traffic state inside the link
    print(W.analyzer.vehicles_to_pandas())  # Information per vehicle
    print(W.analyzer.link_cumulative_to_pandas())
    print(W.analyzer.gps_like_log_to_pandas())
    print(W.analyzer.vehicle_trip_to_pandas())

    # Save the results to CSV
    W.analyzer.output_data()

    assert True

@pytest.mark.filterwarnings("ignore::UserWarning")
def test_osm_import():
    from uxsim.OSMImporter import OSMImporter

    W = World(
        name="",
        deltan=5,
        tmax=7200,
        print_mode=1, save_mode=1, show_mode=0, 
        random_seed=0
    )

    #Tokyo highway
    nodes, links = OSMImporter.import_osm_data(north=35.817, south=35.570, east=139.881, west=139.583, custom_filter='["highway"~"motorway"]')
    nodes, links = OSMImporter.osm_network_postprocessing(nodes, links, node_merge_threshold=0.005, node_merge_iteration=5, enforce_bidirectional=True)  # merge threshold distance: 0.005 degree ~= 500 m. `enforce_bidirectional` makes all links bidirectional, so that network is not fragmented (but the original network topology is not preserved rigorously).
    OSMImporter.osm_network_visualize(nodes, links, show_link_name=0, show_mode=0, save_mode=0)
    OSMImporter.osm_network_to_World(W, nodes, links, default_jam_density=0.2, coef_degree_to_meter=111000)
    W.show_network()
  
    # set demand to central Tokyo from surroundings
    W.adddemand_area2area(139.70, 35.60, 0, 139.75, 35.68, 0.05, 0, 3600, volume=5000)
    W.adddemand_area2area(139.65, 35.70, 0, 139.75, 35.68, 0.05, 0, 3600, volume=5000)
    W.adddemand_area2area(139.75, 35.75, 0, 139.75, 35.68, 0.05, 0, 3600, volume=5000)
    W.adddemand_area2area(139.85, 35.70, 0, 139.75, 35.68, 0.05, 0, 3600, volume=5000)

    W.exec_simulation()

    assert equal_tolerance(len(W.VEHICLES), 5000*4/5)

def test_readme():
    import matplotlib
    matplotlib.use('Agg')
    
    import re, requests
    url = "https://raw.githubusercontent.com/toruseo/UXsim/main/README.md"

    for _ in range(5):
        try:
            response = requests.get(url)
            content = response.text
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout, requests.exceptions.RequestException):
            continue
        break
    else:
        assert True
        return True # Skip the test if the content cannot be retrieved

    pattern = re.compile(r'```python\n(.*?)```', re.DOTALL)
    code_blocks = pattern.findall(content)

    print(code_blocks)

    for code in code_blocks:
        print(code)
        exec(code)

    assert True


def test_scenario_write_and_read():
    ##################################
    # Iter 1
    ##################################

    W = World(
        name="",
        deltan=10,
        tmax=3600,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=42
    )

    imax = 4
    jmax = 4
    nodes, links = generate_grid_network(W, imax, jmax, length=1000)

    demand_flow = 0.035
    demand_duration = 1800
    for n1 in [(0,j) for j in range(jmax)]:
        for n2 in [(imax-1,j) for j in range(jmax)]:
            W.adddemand(nodes[n2], nodes[n1], 0, demand_duration, demand_flow)
            W.adddemand(nodes[n1], nodes[n2], 0, demand_duration, demand_flow)
    for n1 in [(i,0) for i in range(imax)]:
        for n2 in [(i,jmax-1) for i in range(imax)]:
            W.adddemand(nodes[n2], nodes[n1], 0, demand_duration, demand_flow)
            W.adddemand(nodes[n1], nodes[n2], 0, demand_duration, demand_flow)
    W.adddemand_point2point(0.5, 0.5, 2.5, 2.5, 0, 1800, volume=100)
    W.adddemand_area2area(0.5, 0.5, 2, 2.5, 2.5, 2, 0, 1800, volume=100)

    W.save_scenario("out/test_grid.uxsim_scenario")

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    df1 = W.analyzer.basic_to_pandas()
    print(df1)

    ##################################
    # Iter 2
    ##################################

    W2 = World(
        name="",
        deltan=10,
        tmax=3600,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=42
    )

    W2.load_scenario("out/test_grid.uxsim_scenario")

    W2.exec_simulation()

    W2.analyzer.print_simple_stats()

    df2 = W2.analyzer.basic_to_pandas()
    print(df2)
    
    assert df1["total_travel_time"][0] == df2["total_travel_time"][0]

def test_scenario_write_and_read_areas():
    ##################################
    # Iter 1
    ##################################

    W = World(
        name="",
        deltan=10,
        tmax=4000,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=42,
    )

    n_nodes = 5
    imax = n_nodes
    jmax = n_nodes
    nodes = {}
    for i in range(imax):
        for j in range(jmax):
            nodes[i,j] = W.addNode(f"n{(i,j)}", i, j, flow_capacity=1.6)

    links = {}
    for i in range(imax):
        for j in range(jmax):
            if i != imax-1:
                links[i,j,i+1,j] = W.addLink(f"l{(i,j,i+1,j)}", nodes[i,j], nodes[i+1,j], length=1000)
            if i != 0:
                links[i,j,i-1,j] = W.addLink(f"l{(i,j,i-1,j)}", nodes[i,j], nodes[i-1,j], length=1000)
            if j != jmax-1:
                links[i,j,i,j+1] = W.addLink(f"l{(i,j,i,j+1)}", nodes[i,j], nodes[i,j+1], length=1000)
            if j != 0:
                links[i,j,i,j-1] = W.addLink(f"l{(i,j,i,j-1)}", nodes[i,j], nodes[i,j-1], length=1000)


    areas = {
        "areaN": [nodes[0,0], nodes[0, n_nodes-1]],
        "areaS": [nodes[n_nodes-1,0], nodes[n_nodes-1, n_nodes-1]],
        "areaNW": [nodes[0,0]],
        "areaSE": [nodes[n_nodes-1, n_nodes-1]]
    }

    W.adddemand_nodes2nodes(areas["areaN"], areas["areaS"], 0, 3000, volume=7000)

    W.save_scenario("out/test_area.uxsim_scenario")

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    df1 = W.analyzer.basic_to_pandas()
    print(df1)

    ##################################
    # Iter 2
    ##################################

    W2 = World(
        name="",
        deltan=10,
        tmax=4000,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=42,
    )

    W2.load_scenario("out/test_area.uxsim_scenario")

    W2.exec_simulation()
    W2.analyzer.print_simple_stats()

    df2 = W2.analyzer.basic_to_pandas()
    print(df2)

    assert df1["total_travel_time"][0] == df2["total_travel_time"][0]


def test_world_save():
    W = World(
        name="pickle_test",
        deltan=5, 
        tmax=1200,
        print_mode=0, save_mode=0, show_mode=1, 
        random_seed=0
    )

    # Define the scenario

    W.addNode(name="orig1", x=0, y=0)
    W.addNode("orig2", 0, 2)
    W.addNode("merge", 1, 1)
    W.addNode("dest", 2, 1)

    W.addLink(name="link1", start_node="orig1", end_node="merge", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("link2", "orig2", "merge", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("link3", "merge", "dest", length=1000, free_flow_speed=20, number_of_lanes=1)

    W.adddemand(orig="orig1", dest="dest", t_start=0, t_end=1000, flow=0.45)
    W.adddemand("orig2", "dest", 400, 1000, 0.6)

    W.show_network()

    W.save("test_pickle")

    W2 = pickle.load(open("test_pickle.pkl", "rb"))

    W2.exec_simulation()

    W2.analyzer.print_simple_stats() #no print

    W2.analyzer.network(t=500)

    assert True

def test_world_copy():
    
    W = World(
        name="pickle_test",
        deltan=5, 
        tmax=1200,
        print_mode=0, save_mode=0, show_mode=1, 
        random_seed=0
    )

    W.addNode(name="orig1", x=0, y=0)
    W.addNode("orig2", 0, 2)
    W.addNode("merge", 1, 1)
    W.addNode("dest", 2, 1)

    W.addLink(name="link1", start_node="orig1", end_node="merge", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("link2", "orig2", "merge", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("link3", "merge", "dest", length=1000, free_flow_speed=20, number_of_lanes=1)

    W.adddemand(orig="orig1", dest="dest", t_start=0, t_end=1000, flow=0.45)
    W.adddemand("orig2", "dest", 400, 1000, 0.6)

    W.show_network()

    W2 = W.copy()

    W2.exec_simulation()

    W2.analyzer.print_simple_stats() #no print

    W2.analyzer.network(t=500)


def test_k_shortest_path():
    W = World(
        name="",    # Scenario name
        deltan=5,   # Simulation aggregation unit delta n
        tmax=1200,  # Total simulation time (s)
        print_mode=1, save_mode=1, show_mode=1,    # Various options
        random_seed=0    # Set the random seed
    )

    # Define the scenario

    # 2 - D
    # | \ |
    # O - 1
    #free flow travel time: 
    #   O1D = 2000/20 = 100
    #   O2D = 4000/20 = 200
    #   O12D= 1600/20 = 80


    W.addNode(name="O", x=0, y=0)
    W.addNode(name="1", x=1, y=0)
    W.addNode(name="2", x=0, y=1)
    W.addNode(name="D", x=1, y=1)
    W.addLink("O1", "O", "1", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("1D", "1", "D", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("O2", "O", "2", length=3500, free_flow_speed=20, number_of_lanes=1)
    W.addLink("2D", "2", "D", length=500, free_flow_speed=20, number_of_lanes=1)
    W.addLink("12", "1", "2", length=100, free_flow_speed=20, number_of_lanes=1)
    W.adddemand(orig="O", dest="D", t_start=0, t_end=1000, flow=0.6)

    # Run the simulation to the end
    W.exec_simulation()

    # Print summary of simulation result
    W.analyzer.print_simple_stats()

    df = W.analyzer.link_to_pandas()

    assert df[df["link"]=="O1"]["traffic_volume"].values[0] == 600
    assert df[df["link"]=="O2"]["traffic_volume"].values[0] == 0
    assert df[df["link"]=="12"]["traffic_volume"].values[0] == 600

    assert enumerate_k_shortest_routes(W, "O", "D") == [['O1', '12', '2D']]
    assert enumerate_k_shortest_routes(W, "O", "D", k=3) == [['O1', '12', '2D'], ['O1', '1D'], ['O2', '2D']]
    assert enumerate_k_shortest_routes(W, "O", "D", k=3, return_cost=True) == ([['O1', '12', '2D'], ['O1', '1D'], ['O2', '2D']], [80.0, 100.0, 200.0])

@pytest.mark.flaky(reruns=10)
def test_k_shortest_path_on_t():

    W = World(
        name="",    # Scenario name
        deltan=5,   # Simulation aggregation unit delta n
        tmax=1200,  # Total simulation time (s)
        print_mode=1, save_mode=1, show_mode=1,    # Various options
        random_seed=None    # Set the random seed
    )

    # Define the scenario

    # 2 - D
    # | \ |
    # O - 1
    #free flow travel time: 
    #   O1D = 2000/20 = 100
    #   O2D = 4000/20 = 200
    #   O12D= 1600/20 = 80


    W.addNode(name="O", x=0, y=0)
    W.addNode(name="1", x=1, y=0)
    W.addNode(name="2", x=0, y=1)
    W.addNode(name="D", x=1, y=1)
    W.addLink("O1", "O", "1", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("1D", "1", "D", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("O2", "O", "2", length=3500, free_flow_speed=20, number_of_lanes=1)
    W.addLink("2D", "2", "D", length=500, free_flow_speed=20, number_of_lanes=1)
    W.addLink("12", "1", "2", length=100, free_flow_speed=20, number_of_lanes=1, capacity_out=0.4)
    W.adddemand(orig="O", dest="D", t_start=0, t_end=1000, flow=0.6)

    # Run the simulation to the end
    W.exec_simulation()

    # Print summary of simulation result
    W.analyzer.print_simple_stats()

    df = W.analyzer.link_to_pandas()
    
    assert equal_tolerance(df[df["link"]=="O1"]["traffic_volume"].values[0], 460, rel_tol=0.2)
    assert equal_tolerance(df[df["link"]=="O2"]["traffic_volume"].values[0], 140, rel_tol=0.2)
    assert equal_tolerance(df[df["link"]=="12"]["traffic_volume"].values[0], 325, rel_tol=0.2)

    t = 0
    assert enumerate_k_shortest_routes_on_t(W, "O", "D", t=t) == [['O1', '12', '2D']]
    assert enumerate_k_shortest_routes_on_t(W, "O", "D", t=t, k=3) == [['O1', '12', '2D'], ['O1', '1D'], ['O2', '2D']]
    routes, costs = enumerate_k_shortest_routes_on_t(W, "O", "D", t=t, k=3, return_cost=True)
    assert routes[0] == ['O1', '12', '2D']
    assert equal_tolerance(costs[0], 80.0)

    t = 200
    assert enumerate_k_shortest_routes_on_t(W, "O", "D", t=t) == [['O1', '1D']]
    assert enumerate_k_shortest_routes_on_t(W, "O", "D", t=t, k=3) == [['O1', '1D'], ['O1', '12', '2D'], ['O2', '2D']]
    routes, costs = enumerate_k_shortest_routes_on_t(W, "O", "D", t=t, k=3, return_cost=True)
    assert routes[0] == ['O1', '1D']
    assert equal_tolerance(costs[0], 131.8181818181818)

    t = 400
    assert enumerate_k_shortest_routes_on_t(W, "O", "D", t=t) == [['O2', '2D']]
    assert enumerate_k_shortest_routes_on_t(W, "O", "D", t=t, k=3) == [['O2', '2D'], ['O1', '1D'], ['O1', '12', '2D']]
    routes, costs = enumerate_k_shortest_routes_on_t(W, "O", "D", t=t, k=3, return_cost=True)
    assert routes[0] == ['O2', '2D']
    assert equal_tolerance(costs[0], 200.0)

@pytest.mark.flaky(reruns=10)
def test_shortest_path_costs():
    W = World(
        name="",
        deltan=5,
        tmax=1200,
        duo_update_time=600,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=None
    )

    orig = W.addNode("orig", 0, 0)
    W.addNode("mid1", 1, 1)
    W.addNode("mid2", 1, -1)
    dest = W.addNode("dest", 2, 0)
    W.addLink("link01", "orig", "mid1", length=1000, free_flow_speed=20, number_of_lanes=2)
    W.addLink("link02", "orig", "mid2", length=1000, free_flow_speed=20, number_of_lanes=2)
    W.addLink("link13", "mid1", "dest", length=2000, free_flow_speed=20, number_of_lanes=2)
    W.addLink("link23", "mid2", "dest", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.adddemand("orig", "dest", 0, 1000, 1.2)
    W.show_network(figsize=(3,3))

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.cumulative_curves(figsize=(4,2))

    spd = get_shortest_path_distance_between_all_nodes(W)
    assert spd["orig", "dest"] == 2000
    assert spd["orig", "mid1"] == 1000
    assert spd["orig", "mid2"] == 1000
    assert spd["mid1", "dest"] == 2000
    assert spd["mid2", "dest"] == 1000
    assert spd["dest", "orig"] == np.inf
    assert spd[orig, dest] == 2000
    assert spd[dest, orig] == np.inf

    spd = get_shortest_path_distance_between_all_nodes(W, return_matrix=True)
    assert spd[0, 3] == 2000
    assert spd[3, 0] == np.inf

    spt = get_shortest_path_instantaneous_travel_time_between_all_nodes(W)
    assert equal_tolerance(spt["orig", "dest"], 150)
    assert equal_tolerance(spt["orig", "mid1"], 50, rel_tol=0.2)
    assert equal_tolerance(spt["orig", "mid2"], 150)
    assert equal_tolerance(spt["mid1", "dest"], 100)
    assert equal_tolerance(spt["mid2", "dest"], 50, rel_tol=0.2)
    assert spt["dest", "orig"] == np.inf
    assert equal_tolerance(spt[orig, dest], 150)
    assert spt[dest, orig] == np.inf

    spt = get_shortest_path_instantaneous_travel_time_between_all_nodes(W, return_matrix=True)
    assert equal_tolerance(spt[0, 3], 150)
    assert spt[3, 0] == np.inf

    spt0 = get_shortest_path_instantaneous_travel_time_between_all_nodes_on_t(W,0)
    assert equal_tolerance(spt0["orig", "dest"], 100)
    assert equal_tolerance(spt0["orig", "mid1"], 50, rel_tol=0.2)
    assert equal_tolerance(spt0["orig", "mid2"], 50, rel_tol=0.2)
    assert equal_tolerance(spt0["mid1", "dest"], 100)
    assert equal_tolerance(spt0["mid2", "dest"], 50, rel_tol=0.2)

    spt200, t = get_shortest_path_instantaneous_travel_time_between_all_nodes_on_t(W, 200, return_time=True)
    assert spt0 == spt200
    assert t == 0

    spt600 = get_shortest_path_instantaneous_travel_time_between_all_nodes_on_t(W,600)
    assert equal_tolerance(spt600["orig", "dest"], 150)
    assert equal_tolerance(spt600["orig", "mid1"], 50, rel_tol=0.2)
    assert equal_tolerance(spt600["orig", "mid2"], 150)
    assert equal_tolerance(spt600["mid1", "dest"], 100)
    assert equal_tolerance(spt600["mid2", "dest"], 50, rel_tol=0.2)

    spt1200 = get_shortest_path_instantaneous_travel_time_between_all_nodes_on_t(W,1200)
    assert spt600 == spt1200

    spt600_mat = get_shortest_path_instantaneous_travel_time_between_all_nodes_on_t(W, 600, return_matrix=True)
    assert equal_tolerance(spt600_mat[0, 3], 150)
    assert spt600_mat[3, 0] == np.inf

def test_util_catch_exceptions_and_warn():
    with pytest.warns(UserWarning, match=r".*network().*"):
        W = World(
            name="",    # Scenario name
            deltan=5,   # Simulation aggregation unit delta n
            tmax=1200,  # Total simulation time (s)
            print_mode=1, save_mode=1, show_mode=0,    # Various options
            random_seed=0    # Set the random seed
        )

        # Define the scenario
        ## Create nodes
        W.addNode(name="orig1", x=0, y=0)
        W.addNode("orig2", 0, 2)
        W.addNode("merge", 1, 1)
        W.addNode("dest", 2, 1)
        ## Create links between nodes
        W.addLink(name="link1", start_node="orig1", end_node="merge",
                length=1000, free_flow_speed=20, number_of_lanes=1)
        W.addLink("link2", "orig2", "merge", length=1000, free_flow_speed=20, number_of_lanes=1)
        W.addLink("link3", "merge", "dest", length=1000, free_flow_speed=20, number_of_lanes=1)
        ## Create OD traffic demand between nodes
        W.adddemand(orig="orig1", dest="dest", t_start=0, t_end=1000, flow=0.45)
        W.adddemand("orig2", "dest", 400, 1000, 0.6)

        # Run the simulation to the end
        W.exec_simulation()

        # Print summary of simulation result
        W.analyzer.print_simple_stats()

        # Visualize snapshots of network traffic state for several timesteps
        W.analyzer.network()

    assert True

def test_util_print_columns():
    W = World(
        name="",    # Scenario name
        deltan=5,   # Simulation aggregation unit delta n
        tmax=1200,  # Total simulation time (s)
        print_mode=1, save_mode=1, show_mode=0,    # Various options
        random_seed=0    # Set the random seed
    )

    # Define the scenario
    ## Create nodes
    W.addNode(name="orig1", x=0, y=0)
    W.addNode("orig2", 0, 2)
    W.addNode("merge", 1, 1)
    W.addNode("dest", 2, 1)
    ## Create links between nodes
    W.addLink(name="link1", start_node="orig1", end_node="merge",
            length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("link2", "orig2", "merge", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("link3", "merge", "dest", length=1000, free_flow_speed=20, number_of_lanes=1)
    ## Create OD traffic demand between nodes
    W.adddemand(orig="orig1", dest="dest", t_start=0, t_end=1000, flow=0.45)
    W.adddemand("orig2", "dest", 400, 1000, 0.6)

    # Run the simulation to the end
    W.exec_simulation()

    # Print summary of simulation result
    W.analyzer.print_simple_stats()

    print_columns(W.VEHICLES["0"].log_t, W.VEHICLES["0"].log_x, W.VEHICLES["0"].log_v)
    print_columns(W.VEHICLES["0"].log_t, W.VEHICLES["0"].log_x, W.VEHICLES["0"].log_v, W.VEHICLES["1"].log_t, W.VEHICLES["1"].log_x, W.VEHICLES["1"].log_v)

    assert True
    
def test_printtry():
    lis = [1,2,3]
    printtry(lambda: (lis[0]))
    printtry(lambda: (lis[10]))
    assert True

def test_area2area_demand_and_stats():
    W = World(
        name="",
        deltan=10,
        tmax=4000,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=None,
    )

    n_nodes = 5
    imax = n_nodes
    jmax = n_nodes
    nodes = {}
    for i in range(imax):
        for j in range(jmax):
            nodes[i,j] = W.addNode(f"n{(i,j)}", i, j, flow_capacity=1.6)

    links = {}
    for i in range(imax):
        for j in range(jmax):
            if i != imax-1:
                links[i,j,i+1,j] = W.addLink(f"l{(i,j,i+1,j)}", nodes[i,j], nodes[i+1,j], length=1000)
            if i != 0:
                links[i,j,i-1,j] = W.addLink(f"l{(i,j,i-1,j)}", nodes[i,j], nodes[i-1,j], length=1000)
            if j != jmax-1:
                links[i,j,i,j+1] = W.addLink(f"l{(i,j,i,j+1)}", nodes[i,j], nodes[i,j+1], length=1000)
            if j != 0:
                links[i,j,i,j-1] = W.addLink(f"l{(i,j,i,j-1)}", nodes[i,j], nodes[i,j-1], length=1000)


    areas = {
        "areaN": [nodes[0,0], nodes[0, n_nodes-1]],
        "areaS": [nodes[n_nodes-1,0], nodes[n_nodes-1, n_nodes-1]],
        "areaNW": [nodes[0,0]],
        "areaSE": [nodes[n_nodes-1, n_nodes-1]]
    }

    W.adddemand_nodes2nodes(areas["areaN"], areas["areaS"], 0, 3000, volume=7000)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    df = W.analyzer.areas2areas_to_pandas(areas.values(), list(areas.keys()))
    print(df)

    print(W.analyzer.areas2areas_to_pandas(areas.values()))

    assert W.analyzer.trip_all == 7000
    assert df["total_trips"][(df["origin_area"] == "areaN") & (df["destination_area"] == "areaS")].values[0] == 7000
    assert df["average_free_travel_time"][(df["origin_area"] == "areaN") & (df["destination_area"] == "areaS")].values[0] == 300.0
    assert df["average_shortest_distance"][(df["origin_area"] == "areaN") & (df["destination_area"] == "areaS")].values[0] == 6000.0
    assert df["total_trips"][(df["origin_area"] == "areaNW") & (df["destination_area"] == "areaSE")].values[0] == 1750
    assert df["average_free_travel_time"][(df["origin_area"] == "areaNW") & (df["destination_area"] == "areaSE")].values[0] == 400.0
    assert df["average_shortest_distance"][(df["origin_area"] == "areaNW") & (df["destination_area"] == "areaSE")].values[0] == 8000.0

def test_adddemand_area2area2_nodes2nodes2():
    W = World(
        name="",
        deltan=5,
        tmax=7200,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=0
    )

    # scenario
    #automated network generation
    #deploy nodes as an imax x jmax grid
    imax = 5
    jmax = 5
    nodes = {}
    for i in range(imax):
        for j in range(jmax):
            nodes[i,j] = W.addNode(f"n{(i,j)}", i, j)

    #create links between neighborhood nodes
    links = {}
    for i in range(imax):
        for j in range(jmax):
            if i != imax-1:
                links[i,j,i+1,j] = W.addLink(f"l{(i,j,i+1,j)}", nodes[i,j], nodes[i+1,j], length=1000, free_flow_speed=20, jam_density=0.2)
            if i != 0:
                links[i,j,i-1,j] = W.addLink(f"l{(i,j,i-1,j)}", nodes[i,j], nodes[i-1,j], length=1000, free_flow_speed=20, jam_density=0.2)
            if j != jmax-1:
                links[i,j,i,j+1] = W.addLink(f"l{(i,j,i,j+1)}", nodes[i,j], nodes[i,j+1], length=1000, free_flow_speed=20, jam_density=0.2)
            if j != 0:
                links[i,j,i,j-1] = W.addLink(f"l{(i,j,i,j-1)}", nodes[i,j], nodes[i,j-1], length=1000, free_flow_speed=20, jam_density=0.2)

    W.adddemand_nodes2nodes2(nodes.values(), nodes.values(), 0, 3600, volume=5000)
    W.adddemand_area2area2(0, 0, 1.1, 5, 5, 1.1, 0, 3600, volume=5000)

    W.finalize_scenario()

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    assert equal_tolerance(10000, len(W.VEHICLES)*W.DELTAT)

def test_adddemand_area2area2_nodes2nodes2_scenario():
    # simulation world
    W = World(
        name="",
        deltan=5,
        tmax=7200,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=0
    )

    # scenario
    #automated network generation
    #deploy nodes as an imax x jmax grid
    imax = 5
    jmax = 5
    nodes = {}
    for i in range(imax):
        for j in range(jmax):
            nodes[i,j] = W.addNode(f"n{(i,j)}", i, j)

    #create links between neighborhood nodes
    links = {}
    for i in range(imax):
        for j in range(jmax):
            if i != imax-1:
                links[i,j,i+1,j] = W.addLink(f"l{(i,j,i+1,j)}", nodes[i,j], nodes[i+1,j], length=1000, free_flow_speed=20, jam_density=0.2)
            if i != 0:
                links[i,j,i-1,j] = W.addLink(f"l{(i,j,i-1,j)}", nodes[i,j], nodes[i-1,j], length=1000, free_flow_speed=20, jam_density=0.2)
            if j != jmax-1:
                links[i,j,i,j+1] = W.addLink(f"l{(i,j,i,j+1)}", nodes[i,j], nodes[i,j+1], length=1000, free_flow_speed=20, jam_density=0.2)
            if j != 0:
                links[i,j,i,j-1] = W.addLink(f"l{(i,j,i,j-1)}", nodes[i,j], nodes[i,j-1], length=1000, free_flow_speed=20, jam_density=0.2)

    W.adddemand_nodes2nodes2(nodes.values(), nodes.values(), 0, 3600, volume=5000)
    W.adddemand_area2area2(0, 0, 1.1, 5, 5, 1.1, 0, 3600, volume=5000)

    W.save_scenario("out/demandtest.uxsim_scenario")

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    #W.analyzer.network_fancy(animation_speed_inverse=15, sample_ratio=0.2, interval=5, trace_length=10,  figsize=6, antialiasing=False)
    #display_image_in_notebook("out/anim_network_fancy.gif") 


    ### ITER2

    W2 = World(
        name="",
        deltan=5,
        tmax=7200,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=0
    )

    W2.load_scenario("out/demandtest.uxsim_scenario")

    W2.exec_simulation()
    W2.analyzer.print_simple_stats()

    assert W.analyzer.average_travel_time == W2.analyzer.average_travel_time

@pytest.mark.flaky(reruns=10)
def test_area_stats():

    rec_volume_areaN = []
    rec_volume_areaS = []
    rec_ttt_areaN = []
    rec_delay_areaN = []

    for i in range(10):
        W = World(
            name="",
            deltan=10,
            tmax=3000,
            print_mode=1, save_mode=1, show_mode=0,
            random_seed=None,
        )

        n_nodes = 4
        imax = n_nodes
        jmax = n_nodes
        nodes = {}
        for i in range(imax):
            for j in range(jmax):
                nodes[i,j] = W.addNode(f"n{(i,j)}", i, j, flow_capacity=1.6)

        links = {}
        for i in range(imax):
            for j in range(jmax):
                if i != imax-1:
                    links[i,j,i+1,j] = W.addLink(f"l{(i,j,i+1,j)}", nodes[i,j], nodes[i+1,j], length=1000)
                if i != 0:
                    links[i,j,i-1,j] = W.addLink(f"l{(i,j,i-1,j)}", nodes[i,j], nodes[i-1,j], length=1000)
                if j != jmax-1:
                    links[i,j,i,j+1] = W.addLink(f"l{(i,j,i,j+1)}", nodes[i,j], nodes[i,j+1], length=1000)
                if j != 0:
                    links[i,j,i,j-1] = W.addLink(f"l{(i,j,i,j-1)}", nodes[i,j], nodes[i,j-1], length=1000)


        area_dict = {
            "areaN": [nodes[0,i] for i in range(n_nodes)],
            "areaS": [nodes[n_nodes-1,i] for i in range(n_nodes)],
            "areaNW": [nodes[0,0]],
            "areaSE": [nodes[n_nodes-1, n_nodes-1]]
        }

        W.adddemand_nodes2nodes(area_dict["areaN"], area_dict["areaS"], 0, 3000, volume=7000)

        W.exec_simulation()
        W.analyzer.print_simple_stats()

        df = W.analyzer.area_to_pandas(list(area_dict.values()), list(area_dict.keys()), border_include=True)
        print(df)

        rec_volume_areaN.append(df["traffic_volume"][df["area"] == "areaN"].values[0])
        rec_volume_areaS.append(df["traffic_volume"][df["area"] == "areaS"].values[0])
        rec_ttt_areaN.append(df["total_travel_time"][df["area"] == "areaN"].values[0])
        rec_delay_areaN.append(df["average_delay"][df["area"] == "areaN"].values[0])

    assert equal_tolerance(average(rec_volume_areaN), 6880)
    assert equal_tolerance(average(rec_volume_areaS), 6380)
    assert equal_tolerance(average(rec_ttt_areaN), 840000)
    assert equal_tolerance(average(rec_delay_areaN), 0.77, abs_tol=0.1)

@pytest.mark.flaky(reruns=10)
def test_vehicle_group_stats():
    W = World(
        name="",
        deltan=10,
        tmax=4000,
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=None,
    )

    n_nodes = 4
    imax = n_nodes
    jmax = n_nodes
    nodes = {}
    for i in range(imax):
        for j in range(jmax):
            nodes[i,j] = W.addNode(f"n{(i,j)}", i, j, flow_capacity=1.6)

    links = {}
    for i in range(imax):
        for j in range(jmax):
            if i != imax-1:
                links[i,j,i+1,j] = W.addLink(f"l{(i,j,i+1,j)}", nodes[i,j], nodes[i+1,j], length=1000)
            if i != 0:
                links[i,j,i-1,j] = W.addLink(f"l{(i,j,i-1,j)}", nodes[i,j], nodes[i-1,j], length=1000)
            if j != jmax-1:
                links[i,j,i,j+1] = W.addLink(f"l{(i,j,i,j+1)}", nodes[i,j], nodes[i,j+1], length=1000)
            if j != 0:
                links[i,j,i,j-1] = W.addLink(f"l{(i,j,i,j-1)}", nodes[i,j], nodes[i,j-1], length=1000)


    areas = {
        "areaN": [nodes[0,0], nodes[0, n_nodes-1]],
        "areaS": [nodes[n_nodes-1,0], nodes[n_nodes-1, n_nodes-1]],
        "areaNW": [nodes[0,0]],
        "areaSE": [nodes[n_nodes-1, n_nodes-1]]
    }

    W.adddemand_nodes2nodes(areas["areaN"], areas["areaS"], 0, 3000, volume=7000)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    dt = 500
    group_dict = {}
    for t in range(0, W.TMAX, dt):
        group_dict[f"depart_t={t}"] = [veh for veh in W.VEHICLES.values() if t <= veh.departure_time_in_second < t+dt]

    df = W.analyzer.vehicle_groups_to_pandas(list(group_dict.values()), list(group_dict.keys()))
    print(df)

    assert df["average_travel_time"][df["group"]=="depart_t=0"].values[0] < df["average_travel_time"][df["group"]=="depart_t=1500"].values[0]
    assert df["average_delay_ratio"][df["group"]=="depart_t=0"].values[0] < df["average_delay_ratio"][df["group"]=="depart_t=1500"].values[0]
    assert df["average_traveled_distance"][df["group"]=="depart_t=0"].values[0] < df["average_traveled_distance"][df["group"]=="depart_t=1500"].values[0]
    assert df["average_detour_ratio"][df["group"]=="depart_t=0"].values[0] < df["average_detour_ratio"][df["group"]=="depart_t=1500"].values[0]
    assert df["average_speed"][df["group"]=="depart_t=0"].values[0] > df["average_speed"][df["group"]=="depart_t=1500"].values[0]

def test_change_jam_density():
    W = World(
        name="",
        deltan=5,
        tmax=1200,
        print_mode=1, save_mode=0, show_mode=1,
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid1", 0, 2)
    W.addNode("mid2", 1, 1, signal=[240, 240])
    W.addNode("dest", 2, 1)
    W.addLink("link1", "orig", "mid1", length=1000, free_flow_speed=20, number_of_lanes=1)
    link2 = W.addLink("link2", "mid1", "mid2", length=400, free_flow_speed=20, jam_density=0.2, number_of_lanes=1, signal_group=0)
    W.addLink("link3", "mid2", "dest", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.adddemand("orig", "dest", 0, 1200, 0.4)

    check_flag = 0
    while W.check_simulation_ongoing():
        W.exec_simulation(duration_t2=30)
        print(W.TIME, link2.delta_per_lane, link2.num_vehicles, link2.num_vehicles_queue)
        if W.TIME > 600:
            link2.change_jam_density(0.1)
            
        if W.TIME == 180:
            assert equal_tolerance(link2.num_vehicles, 10)
            assert equal_tolerance(link2.num_vehicles_queue, 0)
            check_flag += 1
        if W.TIME == 450:
            assert equal_tolerance(link2.num_vehicles, 400*0.2)
            assert equal_tolerance(link2.num_vehicles_queue, 400*0.2)
            check_flag += 1
        if W.TIME == 540:
            assert equal_tolerance(link2.num_vehicles, 35)
            assert equal_tolerance(link2.num_vehicles_queue, 25)
            check_flag += 1
        if W.TIME == 900:
            assert equal_tolerance(link2.num_vehicles, 400*0.1)
            assert equal_tolerance(link2.num_vehicles_queue, 400*0.1)
            check_flag += 1

    assert check_flag == 4

    W.analyzer.print_simple_stats()

    #W.analyzer.time_space_diagram_traj_links(["link1","link2","link3"])

def test_user_functions():
    # define custom user_functions

    def world_user_function(W):
        #print the current time and the average speed of traveling vehicles in the World
        if W.T%20 == 0:
            if len(W.VEHICLES_RUNNING):
                print("World state:", W.TIME, np.average([veh.v for veh in W.VEHICLES_RUNNING.values()]))
            else:
                print("World state:", W.TIME)

    def node_user_function(node):
        #print the numnber of incoming vehicles to node
        if node.W.T%20 == 0:
            print(f"Node {node.name} state:", len(node.incoming_vehicles))

    def link_user_function(link):
        #print the numnber of vehicles on link and their average speed, and record the speed
        if link.W.T%20 == 0:
            if len(link.vehicles):
                print(f"Link {link.name} state:", len(link.vehicles), np.average([veh.v for veh in link.vehicles]))
                link.user_attribute["speed_record"].append(np.average([veh.v for veh in link.vehicles]))
            else:
                print(f"Link {link.name} state:", len(link.vehicles))


    def veh_user_function(veh):
        #print when the vehicle stated or ended its trip
        if veh.state != "home" and veh.user_attrubute["state_old"] == "home":
            print(f"Vehicle {veh.name} started trip on time {veh.W.TIME}")        
        if veh.state in ["end", "abort"] and veh.user_attrubute["state_old"] not in ["end", "abort"]:
            print(f"Vehicle {veh.name} ended trip on time {veh.W.TIME}")
        veh.user_attrubute["state_old"] = veh.state

    # define uxsim scenario

    W = World(
        name="",
        deltan=5,
        tmax=1200, 
        print_mode=1, save_mode=1, show_mode=0,
        random_seed=0,
        user_function=world_user_function
    )

    W.addNode("orig1", 0, 0)
    W.addNode("orig2", 0, 2)
    W.addNode("merge", 1, 1, user_function=node_user_function)
    W.addNode("dest", 2, 1)
    W.addLink("link1", "orig1", "merge", length=1000, free_flow_speed=20, number_of_lanes=1, user_function=link_user_function, user_attribute={"speed_record":[]})
    W.addLink("link2", "orig2", "merge", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink("link3", "merge", "dest", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.adddemand("orig1", "dest", 0, 1000, 0.45)
    W.adddemand("orig2", "dest", 400, 1000, 0.6)

    W.VEHICLES["10"].user_attrubute = {"state_old":None}
    W.VEHICLES["10"].user_function = veh_user_function
    W.VEHICLES["120"].user_attrubute = {"state_old":None}
    W.VEHICLES["120"].user_function = veh_user_function

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    # prosess data collected by user_function

    link1 = W.get_link("link1")
    print("average speed of link1:", np.nanmean(link1.user_attribute["speed_record"]))

    assert equal_tolerance(np.nanmean(link1.user_attribute["speed_record"]), 15.89)

def test_reduce_memory_delete_vehicle_route_pref():
    W = World(
        name="",
        deltan=10,
        tmax=3000,
        print_mode=1, save_mode=1, show_mode=0,
        reduce_memory_delete_vehicle_route_pref=False,
        random_seed=0,
    )

    n_nodes = 4
    imax = n_nodes
    jmax = n_nodes
    nodes = {}
    for i in range(imax):
        for j in range(jmax):
            nodes[i,j] = W.addNode(f"n{(i,j)}", i, j, flow_capacity=1.6)

    links = {}
    for i in range(imax):
        for j in range(jmax):
            if i != imax-1:
                links[i,j,i+1,j] = W.addLink(f"l{(i,j,i+1,j)}", nodes[i,j], nodes[i+1,j], length=1000)
            if i != 0:
                links[i,j,i-1,j] = W.addLink(f"l{(i,j,i-1,j)}", nodes[i,j], nodes[i-1,j], length=1000)
            if j != jmax-1:
                links[i,j,i,j+1] = W.addLink(f"l{(i,j,i,j+1)}", nodes[i,j], nodes[i,j+1], length=1000)
            if j != 0:
                links[i,j,i,j-1] = W.addLink(f"l{(i,j,i,j-1)}", nodes[i,j], nodes[i,j-1], length=1000)


    area_dict = {
        "areaN": [nodes[0,i] for i in range(n_nodes)],
        "areaS": [nodes[n_nodes-1,i] for i in range(n_nodes)],
        "areaNW": [nodes[0,0]],
        "areaSE": [nodes[n_nodes-1, n_nodes-1]]
    }

    W.adddemand_nodes2nodes(area_dict["areaN"], area_dict["areaS"], 0, 3000, volume=7000)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    df1 = W.analyzer.area_to_pandas(list(area_dict.values()), list(area_dict.keys()), border_include=True)
    print(df1)

    W = World(
        name="",
        deltan=10,
        tmax=3000,
        print_mode=1, save_mode=1, show_mode=0,
        reduce_memory_delete_vehicle_route_pref=True,
        random_seed=0,
    )

    n_nodes = 4
    imax = n_nodes
    jmax = n_nodes
    nodes = {}
    for i in range(imax):
        for j in range(jmax):
            nodes[i,j] = W.addNode(f"n{(i,j)}", i, j, flow_capacity=1.6)

    links = {}
    for i in range(imax):
        for j in range(jmax):
            if i != imax-1:
                links[i,j,i+1,j] = W.addLink(f"l{(i,j,i+1,j)}", nodes[i,j], nodes[i+1,j], length=1000)
            if i != 0:
                links[i,j,i-1,j] = W.addLink(f"l{(i,j,i-1,j)}", nodes[i,j], nodes[i-1,j], length=1000)
            if j != jmax-1:
                links[i,j,i,j+1] = W.addLink(f"l{(i,j,i,j+1)}", nodes[i,j], nodes[i,j+1], length=1000)
            if j != 0:
                links[i,j,i,j-1] = W.addLink(f"l{(i,j,i,j-1)}", nodes[i,j], nodes[i,j-1], length=1000)


    area_dict = {
        "areaN": [nodes[0,i] for i in range(n_nodes)],
        "areaS": [nodes[n_nodes-1,i] for i in range(n_nodes)],
        "areaNW": [nodes[0,0]],
        "areaSE": [nodes[n_nodes-1, n_nodes-1]]
    }

    W.adddemand_nodes2nodes(area_dict["areaN"], area_dict["areaS"], 0, 3000, volume=7000)

    W.exec_simulation()
    W.analyzer.print_simple_stats()

    df2 = W.analyzer.area_to_pandas(list(area_dict.values()), list(area_dict.keys()), border_include=True)
    print(df2)

    assert df1["total_travel_time"][0] == df2["total_travel_time"][0]

def test_route_definition():
    
    import matplotlib
    matplotlib.use('Agg')
    
    W = World(
        name="",
        deltan=5,
        tmax=4000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=42
    )

    # scenario
    W.addNode("0orig", 0, 0)
    W.addNode("1orig", 0, 0)
    W.addNode("2dummy", 2, 0)
    W.addNode("3dummy", 2.2, 0)
    W.addNode("4dest", 3, 0)
    W.addLink("link01", "0orig", "1orig", length=500, free_flow_speed=20, jam_density=0.2)
    W.addLink("link12", "1orig", "2dummy", length=2000, free_flow_speed=20, jam_density=0.2)
    W.addLink("link24", "2dummy", "4dest", length=2000, free_flow_speed=20, jam_density=0.2, capacity_in=0.6)
    W.addLink("link13", "1orig", "3dummy", length=2000, free_flow_speed=15, jam_density=0.2)
    W.addLink("link34", "3dummy", "4dest", length=2000, free_flow_speed=15, jam_density=0.2)
    W.adddemand("0orig", "4dest", 0, 3000, 0.51111111)
    W.adddemand("0orig", "4dest", 1800, 3000, 0.5)


    r1 = W.defRoute(["link01", "link12", "link24"])
    r2 = W.defRoute(["link01", "link13", "link34"])

    # simulation
    W.exec_simulation()

    # results
    W.analyzer.print_simple_stats()
    W.analyzer.time_space_diagram_traj_links(r1.links)
    W.analyzer.time_space_diagram_traj_links(r2.links)

    assert True

def test_route_actual_travel_time():
    W = World(
        name="",
        deltan=5,
        tmax=4000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=42
    )

    # scenario
    W.addNode("0orig", 0, 0)
    W.addNode("1orig", 0, 0)
    W.addNode("2dummy", 2, 0)
    W.addNode("3dummy", 2.2, 0)
    W.addNode("4dest", 3, 0)
    W.addLink("link01", "0orig", "1orig", length=500, free_flow_speed=20, jam_density=0.2)
    W.addLink("link12", "1orig", "2dummy", length=2000, free_flow_speed=20, jam_density=0.2)
    W.addLink("link24", "2dummy", "4dest", length=2000, free_flow_speed=20, jam_density=0.2, capacity_in=0.6)
    W.addLink("link13", "1orig", "3dummy", length=2000, free_flow_speed=15, jam_density=0.2)
    W.addLink("link34", "3dummy", "4dest", length=2000, free_flow_speed=15, jam_density=0.2)
    W.adddemand("0orig", "4dest", 0, 3000, 0.51111111)
    W.adddemand("0orig", "4dest", 1800, 3000, 0.5)


    r1 = W.defRoute(["link01", "link12", "link24"])
    r2 = W.defRoute(["link01", "link13", "link34"])

    # simulation
    W.exec_simulation()

    # results
    W.analyzer.print_simple_stats()

    assert equal_tolerance(r1.actual_travel_time(500), (500+2000+2000)/20)
    assert equal_tolerance(r2.actual_travel_time(500), 500/20+(2000+2000)/15)

    assert equal_tolerance(r1.actual_travel_time(2500), 350)
    assert equal_tolerance(r2.actual_travel_time(2500), 300)

    
def test_route_vehicle_methods():
    W = World(
        name="",
        deltan=5,
        tmax=4000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=42
    )

    # scenario
    W.addNode("0orig", 0, 0)
    W.addNode("1orig", 0, 0)
    W.addNode("2dummy", 2, 0)
    W.addNode("3dummy", 2.2, 0)
    W.addNode("4dest", 3, 0)
    W.addLink("link01", "0orig", "1orig", length=500, free_flow_speed=20, jam_density=0.2)
    W.addLink("link12", "1orig", "2dummy", length=2000, free_flow_speed=20, jam_density=0.2)
    W.addLink("link24", "2dummy", "4dest", length=2000, free_flow_speed=20, jam_density=0.2, capacity_in=0.6)
    W.addLink("link13", "1orig", "3dummy", length=2000, free_flow_speed=15, jam_density=0.2)
    W.addLink("link34", "3dummy", "4dest", length=2000, free_flow_speed=15, jam_density=0.2)
    W.adddemand("0orig", "4dest", 0, 3000, 0.51111111)
    W.adddemand("0orig", "4dest", 1800, 3000, 0.5)


    r1 = W.defRoute(["link01", "link12", "link24"])
    r2 = W.defRoute(["link01", "link13", "link34"])

    # simulation
    W.exec_simulation()

    # results
    W.analyzer.print_simple_stats()

    assert W.VEHICLES["0"].traveled_route()[0] == r1
    assert W.VEHICLES["0"].traveled_route()[0] != r2
    
    assert W.VEHICLES["300"].traveled_route()[0] == r2
    assert W.VEHICLES["300"].traveled_route()[1][0] == 3260
    assert W.VEHICLES["300"].traveled_route()[1][-1] == 3550

    tt_from_vehicle_route = W.VEHICLES["300"].traveled_route()[1][-1]-W.VEHICLES["300"].traveled_route()[1][0]
    tt_from_route_by_departure_time = W.VEHICLES["300"].traveled_route()[0].actual_travel_time(W.VEHICLES["300"].traveled_route()[1][0])
    assert equal_tolerance(tt_from_vehicle_route, tt_from_route_by_departure_time)


def test_route_enforce_route_old():
    W = World(
        name="",
        deltan=5,
        tmax=4000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=42
    )

    # scenario
    W.addNode("0orig", 0, 0)
    W.addNode("1orig", 0, 0)
    W.addNode("2dummy", 2, 0)
    W.addNode("3dummy", 2.2, 0)
    W.addNode("4dest", 3, 0)
    W.addLink("link01", "0orig", "1orig", length=500, free_flow_speed=20, jam_density=0.2)
    W.addLink("link12", "1orig", "2dummy", length=2000, free_flow_speed=20, jam_density=0.2)
    W.addLink("link24", "2dummy", "4dest", length=2000, free_flow_speed=20, jam_density=0.2, capacity_in=0.6)
    W.addLink("link13", "1orig", "3dummy", length=2000, free_flow_speed=15, jam_density=0.2)
    W.addLink("link34", "3dummy", "4dest", length=2000, free_flow_speed=15, jam_density=0.2)
    W.adddemand("0orig", "4dest", 0, 3000, 0.51111111)
    W.adddemand("0orig", "4dest", 1800, 3000, 0.5)


    r1 = W.defRoute(["link01", "link12", "link24"])
    r2 = W.defRoute(["link01", "link13", "link34"])

    for veh in W.VEHICLES.values():
        veh.enforce_route(r2)
    
    # simulation
    W.exec_simulation()

    # results
    W.analyzer.print_simple_stats()

    df = W.analyzer.link_to_pandas()
    for l in r2:
        assert df[df["link"]==l.name]["traffic_volume"].values[0] == 2130

def test_route_enforce_route_by_route_object():

    W = World(
        name="looproute",    # Scenario name
        deltan=5,   # Simulation aggregation unit delta n
        tmax=2400,  # Total simulation time (s)
        print_mode=1, save_mode=0, show_mode=1,    # Various options
        random_seed=0    # Set the random seed
    )

    # Define the scenario
    ## Create nodes
    W.addNode(name="O", x=0, y=0)
    W.addNode("D", 2, 0)
    W.addNode("A", 1, 0)
    W.addNode("B", 1.5, 1)
    W.addNode("C", 0.5, 1)
    W.addNode("E", 1, -1)
    ## Create links between nodes
    W.addLink(None, "O", "A", length=500, free_flow_speed=20, number_of_lanes=1)
    W.addLink(None, "A", "D", length=500, free_flow_speed=20, number_of_lanes=1)
    W.addLink(None, "A", "B", length=500, free_flow_speed=20, number_of_lanes=1)
    W.addLink(None, "B", "C", length=500, free_flow_speed=20, number_of_lanes=1)
    W.addLink(None, "C", "A", length=500, free_flow_speed=20, number_of_lanes=1)
    W.addLink(None, "O", "E", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink(None, "E", "D", length=1000, free_flow_speed=20, number_of_lanes=1)
    W.addLink(None, "O", "D", length=500, free_flow_speed=40, number_of_lanes=1)

    ## Create OD traffic demand between nodes
    W.adddemand("O","D", t_start=0, t_end=1000, volume=400)

    #W.show_network()

    route_straight = W.defRoute(["O-A", "A-D"])
    route_detour_loop = W.defRoute(["O-A", "A-B", "B-C", "C-A", "A-D"])
    route_detour_loop_triple = W.defRoute(["O-A", "A-B", "B-C", "C-A", "A-B", "B-C", "C-A", "A-B", "B-C", "C-A", "A-D"])
    route_detour = W.defRoute(["O-E", "E-D"])

    for i,veh in enumerate(W.VEHICLES.values()):
        if i%4 == 0:
            veh.enforce_route(route_straight)
        elif i%4 == 1:
            veh.enforce_route(route_detour)
        elif i%4 == 2:
            veh.enforce_route(route_detour_loop)
        else:
            veh.enforce_route(route_detour_loop_triple)

    # Run the simulation to the end
    W.exec_simulation()

    # Print summary of simulation result
    W.analyzer.print_simple_stats()

    W.analyzer.network_average()
    df = W.analyzer.link_to_pandas()
    print(df)
    
    assert df.query("link == 'O-A'")["traffic_volume"].item() == 300
    assert df.query("link == 'A-D'")["traffic_volume"].item() == 300
    assert df.query("link == 'A-B'")["traffic_volume"].item() == 400
    assert df.query("link == 'B-C'")["traffic_volume"].item() == 400
    assert df.query("link == 'C-A'")["traffic_volume"].item() == 400
    assert df.query("link == 'O-E'")["traffic_volume"].item() == 100
    assert df.query("link == 'E-D'")["traffic_volume"].item() == 100
    assert df.query("link == 'O-D'")["traffic_volume"].item() == 0

    assert W.VEHICLES["0"].traveled_route()[0] == route_straight
    assert W.VEHICLES["1"].traveled_route()[0] == route_detour
    assert W.VEHICLES["2"].traveled_route()[0] == route_detour_loop
    assert W.VEHICLES["3"].traveled_route()[0] == route_detour_loop_triple

def test_construct_time_space_network():
    W = World(
        name="",
        deltan=20,
        tmax=6000,
        print_mode=1, save_mode=1, show_mode=1,
        vehicle_logging_timestep_interval=1, 
        hard_deterministic_mode=False,
        random_seed=42    #fix seed to reproduce random demand 
    )

    W.addNode("1", 0, 1)
    W.addNode("2", 1, 1)
    W.addNode("3", 5, 1)
    W.addNode("4", 0, 0)
    W.addNode("5", 1, 0)
    W.addNode("6", 5, 0)
    W.addNode("7", 6, 0.5)

    W.addLink("highway12", "1", "2", length=1000, number_of_lanes=1, merge_priority=1)
    W.addLink("highway23", "2", "3", length=3000, number_of_lanes=1, merge_priority=1, capacity_out=0.6)
    W.addLink("highway37", "3", "7", length=1000, number_of_lanes=1, merge_priority=1)
    W.addLink("onramp", "5", "2", length=1000, number_of_lanes=1, merge_priority=0.5)
    W.addLink("arterial45", "4", "5", length=1000, free_flow_speed=10, number_of_lanes=2, merge_priority=0.5)
    W.addLink("arterial56", "5", "6", length=3000, free_flow_speed=10, number_of_lanes=2, merge_priority=0.5)
    W.addLink("arterial67", "6", "7", length=1000, free_flow_speed=10, number_of_lanes=2, merge_priority=0.5)

    W.adddemand("1", "7", 0, 3000, 0.3)
    W.adddemand("4", "7", 0, 3000, 0.4*3)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    construct_time_space_network(W)

    assert W.TSN_paths["4", 0]["7", "end"][-2] == ('7', 340)
    assert equal_tolerance(W.TSN_costs["4", 0]["7", "end"], W.TSN_paths["4", 0]["7", "end"][-2][1])

@pytest.mark.flaky(reruns=10)
def test_estimate_congestion_externality_1link():
    def create_world(seed):
        W = World(
            name="aaa",    # Scenario name
            deltan=10,   # Simulation aggregation unit delta n
            tmax=2400,  # Total simulation time (s)
            print_mode=0, save_mode=0, show_mode=0,    # Various options
            random_seed=seed    # Set the random seed
        )

        # Define the scenario
        ## Create nodes
        W.addNode(name="orig1", x=0, y=0)
        W.addNode("mid", 1, 1)
        W.addNode("dest", 2, 1)
        ## Create links between nodes
        link1 = W.addLink(name="link1", start_node="orig1", end_node="mid", length=2000, free_flow_speed=20, number_of_lanes=1, capacity_out=0.8)
        link2 = W.addLink("link3", "mid", "dest", length=1000, free_flow_speed=20, number_of_lanes=1, capacity_in=0.4)

        R1 = uxsim.Route(W, [link1, link2])
        ## Create OD traffic demand between nodes
        W.adddemand(orig="orig1", dest="dest", t_start=0, t_end=1200, flow=0.2)
        W.adddemand(orig="orig1", dest="dest", t_start=300, t_end=600, flow=0.4)

        return W

    dep_ts = []
    acts = []
    ests = []
    for _ in range(50):
        seed = random.randint(0,999999)

        W = create_world(seed)
        W.exec_simulation()
        W.analyzer.print_simple_stats()
        #W.analyzer.time_space_diagram_traj_links(["link1", "link3"])

        key = random.choice(list(W.VEHICLES.keys()))
        veh = W.VEHICLES[key]
        veh_tt = veh.travel_time*W.DELTAN
        route = veh.traveled_route()[0]
        dep_t = veh.log_t_link[1][0]
        ext = estimate_congestion_externality_route(W, route, dep_t)
        ttt1 = W.analyzer.total_travel_time


        W2 = create_world(seed)
        W2.VEHICLES[key].state="abort"
        W2.exec_simulation()
        W2.analyzer.print_simple_stats()
        #W2.analyzer.time_space_diagram_traj_links(["link1", "link3"])
        ttt2 = W2.analyzer.total_travel_time

        #print(f"selected vehicle - dep time {veh.departure_time_in_second}")
        #print("actual ext:", ttt1-ttt2-veh_tt, f"{ttt1}-{ttt2}-{veh_tt}")
        #print("esti. ext: ", ext)

        dep_ts.append(veh.departure_time_in_second)
        acts.append(ttt1-ttt2-veh_tt)
        ests.append(ext)

    # W.analyzer.time_space_diagram_traj_links(["link1", "link3"])

    # figure()
    # plot(dep_ts, acts, "o", label="actual")
    # plot(dep_ts, ests, "x", label="estimated")
    # legend()
    # grid()

    assert uxsim.equal_tolerance(average(acts), average(ests), rel_tol=0.333)


"""
This script tests optional functions in UXsim available for Python 3.10 or later (the requirement for UXsim is Python 3.9).
"""

import pytest
from numpy import *
from uxsim import *

@pytest.mark.filterwarnings("ignore::UserWarning")
def test_osm_import2():
    """
    Requires 3.11 or later.
    This test is affected by OSM update.
    """

    from uxsim.OSMImporter2 import OSMImporter2

    W = World(
        name="", 
        deltan=5,
        tmax=3600,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    # bbox=(139.583, 35.570, 139.881, 35.817)
    # custom_filter='["highway"~"motorway"]' #高速道路

    bbox=(139.638, 35.621, 139.700, 35.600)
    custom_filter='["highway"~"motorway|trunk|primary|secondary|tertiary"]'

    OSMImporter2.osm_network_to_World(W, bbox=bbox, custom_filter=custom_filter)

    W.adddemand_nodes2nodes2(W.NODES, W.NODES, 0, 2000, volume=2.0*len(W.NODES)**2)
    W.exec_simulation()
    W.analyzer.print_simple_stats()

    #W.analyzer.network_average(network_font_size=0)

    assert equal_tolerance(W.analyzer.basic_to_pandas()["average_travel_time"][0], 327, rel_tol=2)


"""
This script tests ResultGUIViewer.
This is for local test only. I don't know how to test it in Github Actions properly.
"""

import pytest

def test_result_gui_viewer():
    import uxsim as UX
    from uxsim.ResultGUIViewer import ResultGUIViewer

    #simple network
    W = UX.World(
        name="",    # Scenario name
        deltan=5,   # Simulation aggregation unit delta n
        tmax=1200,  # Total simulation time (s)
        print_mode=0, save_mode=0, show_mode=0,    # Various options
        random_seed=None    # Set the random seed
    )

    # Define the scenario
    W.addNode("orig1", 0, 0) # Create a node
    W.addNode("orig2", 0, 2)
    W.addNode("merge", 0.5, 1)
    W.addNode("dest", 1, 1)
    W.addLink("link1", "orig1", "merge", length=1000, free_flow_speed=20, jam_density=0.2, merge_priority=0.5) # Create a link
    W.addLink("link2", "orig2", "merge", length=1000, free_flow_speed=20, jam_density=0.2, merge_priority=2)
    W.addLink("link3", "merge", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig1", "dest", 0, 1000, 0.4) # Create OD traffic demand
    W.adddemand("orig2", "dest", 500, 1000, 0.6)


    # Run the simulation to the end
    W.exec_simulation()

    # Print summary of simulation result
    W.analyzer.print_simple_stats()

    app, window = ResultGUIViewer.launch_World_viewer(W, return_app_window=True)

    # Shut down 5 sec later
    from PyQt5.QtCore import QTimer
    import sys
    timer = QTimer()
    timer.timeout.connect(window.close)
    timer.start(1000*5)
    assert True
    with pytest.raises(SystemExit):
        sys.exit(app.exec_())


"""
This script verifies whether UXsim outputs reasonable solutions for a straight road in various configurations.
"""

import pytest
import random
from uxsim import *

"""
default FD:
    u = 20
    kappa = 0.2
    tau = 1
    w = 5
    k^* = 0.04
    q^* = 0.8
"""

def test_1link():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("dest", 1, 1)
    link = W.addLink("link", "orig", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.5)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 250)
    assert equal_tolerance(W.analyzer.trip_completed, 250)
    assert equal_tolerance(W.analyzer.total_travel_time, 12500)
    assert equal_tolerance(W.analyzer.average_travel_time, 50)
    assert equal_tolerance(W.analyzer.average_delay, 0)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link.q_mat[2, 5], 0.5)
    assert equal_tolerance(link.q_mat[7, 5], 0)
    assert equal_tolerance(link.k_mat[2, 5], 0.025)
    assert equal_tolerance(link.k_mat[7, 5], 0)
    assert equal_tolerance(link.v_mat[2, 5], 20)
    assert equal_tolerance(link.v_mat[7, 5], 20)

def test_1link_demand_by_volume():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("dest", 1, 1)
    link = W.addLink("link", "orig", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, volume=0.5*500)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 250)
    assert equal_tolerance(W.analyzer.trip_completed, 250)
    assert equal_tolerance(W.analyzer.total_travel_time, 12500)
    assert equal_tolerance(W.analyzer.average_travel_time, 50)
    assert equal_tolerance(W.analyzer.average_delay, 0)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link.q_mat[2, 5], 0.5)
    assert equal_tolerance(link.q_mat[7, 5], 0)
    assert equal_tolerance(link.k_mat[2, 5], 0.025)
    assert equal_tolerance(link.k_mat[7, 5], 0)
    assert equal_tolerance(link.v_mat[2, 5], 20)
    assert equal_tolerance(link.v_mat[7, 5], 20)

def test_1link_iterative_exec():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("dest", 1, 1)
    link = W.addLink("link", "orig", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.5)

    while W.check_simulation_ongoing():
        W.exec_simulation(duration_t=random.randint(50, 200)) 

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 250)
    assert equal_tolerance(W.analyzer.trip_completed, 250)
    assert equal_tolerance(W.analyzer.total_travel_time, 12500)
    assert equal_tolerance(W.analyzer.average_travel_time, 50)
    assert equal_tolerance(W.analyzer.average_delay, 0)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link.q_mat[2, 5], 0.5)
    assert equal_tolerance(link.q_mat[7, 5], 0)
    assert equal_tolerance(link.k_mat[2, 5], 0.025)
    assert equal_tolerance(link.k_mat[7, 5], 0)
    assert equal_tolerance(link.v_mat[2, 5], 20)
    assert equal_tolerance(link.v_mat[7, 5], 20)

def test_1link_deltan1():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("dest", 1, 1)
    link = W.addLink("link", "orig", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.5)

    W.exec_simulation()

    W.analyzer.print_simple_stats()


    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 250)
    assert equal_tolerance(W.analyzer.trip_completed, 250)
    assert equal_tolerance(W.analyzer.total_travel_time, 12500)
    assert equal_tolerance(W.analyzer.average_travel_time, 50)
    assert equal_tolerance(W.analyzer.average_delay, 0)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link.q_mat[2, 5], 0.5)
    assert equal_tolerance(link.q_mat[7, 5], 0)
    assert equal_tolerance(link.k_mat[2, 5], 0.025)
    assert equal_tolerance(link.k_mat[7, 5], 0)
    assert equal_tolerance(link.v_mat[2, 5], 20)
    assert equal_tolerance(link.v_mat[7, 5], 20)

def test_1link_maxflow():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("dest", 1, 1)
    link = W.addLink("link", "orig", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 2000, 2)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link.q_mat[2, 5], 0.8)
    assert equal_tolerance(link.q_mat[7, 5], 0.8)
    assert equal_tolerance(link.k_mat[2, 5], 0.04)
    assert equal_tolerance(link.k_mat[7, 5], 0.04)
    assert equal_tolerance(link.v_mat[2, 5], 20)
    assert equal_tolerance(link.v_mat[7, 5], 20)    

def test_1link_maxflow_deltan1():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("dest", 1, 1)
    link = W.addLink("link", "orig", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 2000, 2)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link.q_mat[2, 5], 0.8)
    assert equal_tolerance(link.q_mat[7, 5], 0.8)
    assert equal_tolerance(link.k_mat[2, 5], 0.04)
    assert equal_tolerance(link.k_mat[7, 5], 0.04)
    assert equal_tolerance(link.v_mat[2, 5], 20)
    assert equal_tolerance(link.v_mat[7, 5], 20) 

def test_2link():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=10, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.5)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 250)
    assert equal_tolerance(W.analyzer.trip_completed, 250)
    assert equal_tolerance(W.analyzer.total_travel_time, 37500)
    assert equal_tolerance(W.analyzer.average_travel_time, 150)
    assert equal_tolerance(W.analyzer.average_delay, 0)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.5)
    assert equal_tolerance(link1.q_mat[7, 5], 0)
    assert equal_tolerance(link1.k_mat[2, 5], 0.025)
    assert equal_tolerance(link1.k_mat[7, 5], 0)
    assert equal_tolerance(link1.v_mat[2, 5], 20)
    assert equal_tolerance(link1.v_mat[7, 5], 20)
    assert equal_tolerance(link2.q_mat[2, 5], 0.5)
    assert equal_tolerance(link2.q_mat[7, 5], 0)
    assert equal_tolerance(link2.k_mat[2, 5], 0.05)
    assert equal_tolerance(link2.k_mat[7, 5], 0)
    assert equal_tolerance(link2.v_mat[2, 5], 10)
    assert equal_tolerance(link2.v_mat[7, 5], 10)

def test_2link_deltan1():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=10, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.5)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 250)
    assert equal_tolerance(W.analyzer.trip_completed, 250)
    assert equal_tolerance(W.analyzer.total_travel_time, 37500)
    assert equal_tolerance(W.analyzer.average_travel_time, 150)
    assert equal_tolerance(W.analyzer.average_delay, 0)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.5)
    assert equal_tolerance(link1.q_mat[7, 5], 0)
    assert equal_tolerance(link1.k_mat[2, 5], 0.025)
    assert equal_tolerance(link1.k_mat[7, 5], 0)
    assert equal_tolerance(link1.v_mat[2, 5], 20)
    assert equal_tolerance(link1.v_mat[7, 5], 20)
    assert equal_tolerance(link2.q_mat[2, 5], 0.5)
    assert equal_tolerance(link2.q_mat[7, 5], 0)
    assert equal_tolerance(link2.k_mat[2, 5], 0.05)
    assert equal_tolerance(link2.k_mat[7, 5], 0)
    assert equal_tolerance(link2.v_mat[2, 5], 10)
    assert equal_tolerance(link2.v_mat[7, 5], 10)

def test_2link_maxflow():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 2000, 2)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.8)
    assert equal_tolerance(link1.q_mat[7, 5], 0.8)
    assert equal_tolerance(link1.k_mat[2, 5], 0.04)
    assert equal_tolerance(link1.k_mat[7, 5], 0.04)
    assert equal_tolerance(link1.v_mat[2, 5], 20)
    assert equal_tolerance(link1.v_mat[7, 5], 20)
    assert equal_tolerance(link2.q_mat[2, 5], 0.8)
    assert equal_tolerance(link2.q_mat[7, 5], 0.8)
    assert equal_tolerance(link2.k_mat[2, 5], 0.04)
    assert equal_tolerance(link2.k_mat[7, 5], 0.04)
    assert equal_tolerance(link2.v_mat[2, 5], 20)
    assert equal_tolerance(link2.v_mat[7, 5], 20)

def test_2link_maxflow_deltan1():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 2000, 2)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.8)
    assert equal_tolerance(link1.q_mat[7, 5], 0.8)
    assert equal_tolerance(link1.k_mat[2, 5], 0.04)
    assert equal_tolerance(link1.k_mat[7, 5], 0.04)
    assert equal_tolerance(link1.v_mat[2, 5], 20)
    assert equal_tolerance(link1.v_mat[7, 5], 20)
    assert equal_tolerance(link2.q_mat[2, 5], 0.8)
    assert equal_tolerance(link2.q_mat[7, 5], 0.8)
    assert equal_tolerance(link2.k_mat[2, 5], 0.04)
    assert equal_tolerance(link2.k_mat[7, 5], 0.04)
    assert equal_tolerance(link2.v_mat[2, 5], 20)
    assert equal_tolerance(link2.v_mat[7, 5], 20)

def test_2link_bottleneck_due_to_free_flow_speed():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=10, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.8)
    W.adddemand("orig", "dest", 500, 1500, 0.4)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 800)
    assert equal_tolerance(W.analyzer.trip_completed, 800)
    assert equal_tolerance(W.analyzer.total_travel_time, 146000.0)
    assert equal_tolerance(W.analyzer.average_travel_time, 182.5)
    assert equal_tolerance(W.analyzer.average_delay, 32.5)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.6667) #congestion before BN
    assert equal_tolerance(link1.k_mat[2, 5], 0.06667) #congestion before BN
    assert equal_tolerance(link1.v_mat[2, 5], 10) #congestion before BN
    assert equal_tolerance(link1.q_mat[7, 5], 0.4) #freeflow
    assert equal_tolerance(link1.k_mat[7, 5], 0.02) #freeflow
    assert equal_tolerance(link1.v_mat[7, 5], 20) #freeflow
    assert equal_tolerance(link2.q_mat[2, 5], 0.6667) #freeflow after BN
    assert equal_tolerance(link2.k_mat[2, 5], 0.06667) #freeflow after BN
    assert equal_tolerance(link2.v_mat[2, 5], 10) #freeflow after BN
    assert equal_tolerance(link2.q_mat[8, 5], 0.4) #freeflow
    assert equal_tolerance(link2.k_mat[8, 5], 0.04) #freeflow 
    assert equal_tolerance(link2.v_mat[8, 5], 10) #freeflow

def test_2link_bottleneck_due_to_free_flow_speed_deltan1():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=10, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.8)
    W.adddemand("orig", "dest", 500, 1500, 0.4)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 800)
    assert equal_tolerance(W.analyzer.trip_completed, 800)
    assert equal_tolerance(W.analyzer.total_travel_time, 146000.0)
    assert equal_tolerance(W.analyzer.average_travel_time, 182.5)
    assert equal_tolerance(W.analyzer.average_delay, 32.5)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.6667) #congestion before BN
    assert equal_tolerance(link1.k_mat[2, 5], 0.06667) #congestion before BN
    assert equal_tolerance(link1.v_mat[2, 5], 10) #congestion before BN
    assert equal_tolerance(link1.q_mat[7, 5], 0.4) #freeflow
    assert equal_tolerance(link1.k_mat[7, 5], 0.02) #freeflow
    assert equal_tolerance(link1.v_mat[7, 5], 20) #freeflow
    assert equal_tolerance(link2.q_mat[2, 5], 0.6667) #freeflow after BN
    assert equal_tolerance(link2.k_mat[2, 5], 0.06667) #freeflow after BN
    assert equal_tolerance(link2.v_mat[2, 5], 10) #freeflow after BN
    assert equal_tolerance(link2.q_mat[8, 5], 0.4) #freeflow
    assert equal_tolerance(link2.k_mat[8, 5], 0.04) #freeflow 
    assert equal_tolerance(link2.v_mat[8, 5], 10) #freeflow

def test_2link_bottleneck_due_to_capacity_out():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.66666)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.8)
    W.adddemand("orig", "dest", 500, 1500, 0.4)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 800)
    assert equal_tolerance(W.analyzer.trip_completed, 800)
    assert equal_tolerance(W.analyzer.total_travel_time, 104775)
    assert equal_tolerance(W.analyzer.average_travel_time, 131)
    assert equal_tolerance(W.analyzer.average_delay, 31)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.6667) #congestion before BN
    assert equal_tolerance(link1.k_mat[2, 5], 0.06667) #congestion before BN
    assert equal_tolerance(link1.v_mat[2, 5], 10) #congestion before BN
    assert equal_tolerance(link1.q_mat[7, 5], 0.4) #freeflow
    assert equal_tolerance(link1.k_mat[7, 5], 0.02) #freeflow
    assert equal_tolerance(link1.v_mat[7, 5], 20) #freeflow
    assert equal_tolerance(link2.q_mat[2, 5], 0.6667) #freeflow after BN
    assert equal_tolerance(link2.k_mat[2, 5], 0.03333) #freeflow after BN
    assert equal_tolerance(link2.v_mat[2, 5], 20) #freeflow after BN
    assert equal_tolerance(link2.q_mat[8, 5], 0.4) #freeflow
    assert equal_tolerance(link2.k_mat[8, 5], 0.02) #freeflow 
    assert equal_tolerance(link2.v_mat[8, 5], 20) #freeflow

def test_2link_bottleneck_due_to_capacity_out_deltan1():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.66666)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.8)
    W.adddemand("orig", "dest", 500, 1500, 0.4)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 800)
    assert equal_tolerance(W.analyzer.trip_completed, 800)
    assert equal_tolerance(W.analyzer.total_travel_time, 104775)
    assert equal_tolerance(W.analyzer.average_travel_time, 131)
    assert equal_tolerance(W.analyzer.average_delay, 31)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.6667) #congestion before BN
    assert equal_tolerance(link1.k_mat[2, 5], 0.06667) #congestion before BN
    assert equal_tolerance(link1.v_mat[2, 5], 10) #congestion before BN
    assert equal_tolerance(link1.q_mat[7, 5], 0.4) #freeflow
    assert equal_tolerance(link1.k_mat[7, 5], 0.02) #freeflow
    assert equal_tolerance(link1.v_mat[7, 5], 20) #freeflow
    assert equal_tolerance(link2.q_mat[2, 5], 0.6667) #freeflow after BN
    assert equal_tolerance(link2.k_mat[2, 5], 0.03333) #freeflow after BN
    assert equal_tolerance(link2.v_mat[2, 5], 20) #freeflow after BN
    assert equal_tolerance(link2.q_mat[8, 5], 0.4) #freeflow
    assert equal_tolerance(link2.k_mat[8, 5], 0.02) #freeflow 
    assert equal_tolerance(link2.v_mat[8, 5], 20) #freeflow

def test_2link_bottleneck_due_to_jam_density():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.1)
    W.adddemand("orig", "dest", 0, 500, 0.8)
    W.adddemand("orig", "dest", 500, 1500, 0.4)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 800)
    assert equal_tolerance(W.analyzer.trip_completed, 800)
    assert equal_tolerance(W.analyzer.total_travel_time, 106000)
    assert equal_tolerance(W.analyzer.average_travel_time, 132.5)
    assert equal_tolerance(W.analyzer.average_delay, 32.5)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.6667) #congestion before BN
    assert equal_tolerance(link1.k_mat[2, 5], 0.06667) #congestion before BN
    assert equal_tolerance(link1.v_mat[2, 5], 10) #congestion before BN
    assert equal_tolerance(link1.q_mat[7, 5], 0.4) #freeflow
    assert equal_tolerance(link1.k_mat[7, 5], 0.02) #freeflow
    assert equal_tolerance(link1.v_mat[7, 5], 20) #freeflow
    assert equal_tolerance(link2.q_mat[2, 5], 0.6667) #freeflow after BN
    assert equal_tolerance(link2.k_mat[2, 5], 0.03333) #freeflow after BN
    assert equal_tolerance(link2.v_mat[2, 5], 20) #freeflow after BN
    assert equal_tolerance(link2.q_mat[8, 5], 0.4) #freeflow
    assert equal_tolerance(link2.k_mat[8, 5], 0.02) #freeflow 
    assert equal_tolerance(link2.v_mat[8, 5], 20) #freeflow

def test_2link_bottleneck_due_to_node_capacity():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1, flow_capacity=0.666)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.8)
    W.adddemand("orig", "dest", 500, 1500, 0.4)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 800)
    assert equal_tolerance(W.analyzer.trip_completed, 800)
    assert equal_tolerance(W.analyzer.total_travel_time, 106000)
    assert equal_tolerance(W.analyzer.average_travel_time, 132.5)
    assert equal_tolerance(W.analyzer.average_delay, 32.5)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.6667) #congestion before BN
    assert equal_tolerance(link1.k_mat[2, 5], 0.06667) #congestion before BN
    assert equal_tolerance(link1.v_mat[2, 5], 10) #congestion before BN
    assert equal_tolerance(link1.q_mat[7, 5], 0.4) #freeflow
    assert equal_tolerance(link1.k_mat[7, 5], 0.02) #freeflow
    assert equal_tolerance(link1.v_mat[7, 5], 20) #freeflow
    assert equal_tolerance(link2.q_mat[2, 5], 0.6667) #freeflow after BN
    assert equal_tolerance(link2.k_mat[2, 5], 0.03333) #freeflow after BN
    assert equal_tolerance(link2.v_mat[2, 5], 20) #freeflow after BN
    assert equal_tolerance(link2.q_mat[8, 5], 0.4) #freeflow
    assert equal_tolerance(link2.k_mat[8, 5], 0.02) #freeflow 
    assert equal_tolerance(link2.v_mat[8, 5], 20) #freeflow

def test_2link_bottleneck_due_to_node_capacity_deltan1():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1, flow_capacity=0.666)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.8)
    W.adddemand("orig", "dest", 500, 1500, 0.4)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 800)
    assert equal_tolerance(W.analyzer.trip_completed, 800)
    assert equal_tolerance(W.analyzer.total_travel_time, 106000)
    assert equal_tolerance(W.analyzer.average_travel_time, 132.5)
    assert equal_tolerance(W.analyzer.average_delay, 32.5)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.6667) #congestion before BN
    assert equal_tolerance(link1.k_mat[2, 5], 0.06667) #congestion before BN
    assert equal_tolerance(link1.v_mat[2, 5], 10) #congestion before BN
    assert equal_tolerance(link1.q_mat[7, 5], 0.4) #freeflow
    assert equal_tolerance(link1.k_mat[7, 5], 0.02) #freeflow
    assert equal_tolerance(link1.v_mat[7, 5], 20) #freeflow
    assert equal_tolerance(link2.q_mat[2, 5], 0.6667) #freeflow after BN
    assert equal_tolerance(link2.k_mat[2, 5], 0.03333) #freeflow after BN
    assert equal_tolerance(link2.v_mat[2, 5], 20) #freeflow after BN
    assert equal_tolerance(link2.q_mat[8, 5], 0.4) #freeflow
    assert equal_tolerance(link2.k_mat[8, 5], 0.02) #freeflow 
    assert equal_tolerance(link2.v_mat[8, 5], 20) #freeflow

def test_2link_leave_at_middle():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1)
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "mid", 0, 500, 0.666)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 330)
    assert equal_tolerance(W.analyzer.trip_completed, 330)
    assert equal_tolerance(W.analyzer.total_travel_time, 17300)
    assert equal_tolerance(W.analyzer.average_travel_time, 52.4)
    assert equal_tolerance(W.analyzer.average_delay, 2.4)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[2, 5], 0.6667) #congestion before BN
    assert equal_tolerance(link1.k_mat[2, 5], 0.03333) #congestion before BN
    assert equal_tolerance(link1.v_mat[2, 5], 20) #congestion before BN
    assert equal_tolerance(link1.q_mat[7, 5], 0.0) #freeflow
    assert equal_tolerance(link1.k_mat[7, 5], 0.0) #freeflow
    assert equal_tolerance(link1.v_mat[7, 5], 20) #freeflow
    assert equal_tolerance(link2.q_mat[2, 5], 0.0) #freeflow after BN
    assert equal_tolerance(link2.k_mat[2, 5], 0.0) #freeflow after BN
    assert equal_tolerance(link2.v_mat[2, 5], 20) #freeflow after BN
    assert equal_tolerance(link2.q_mat[8, 5], 0.0) #freeflow
    assert equal_tolerance(link2.k_mat[8, 5], 0.0) #freeflow 
    assert equal_tolerance(link2.v_mat[8, 5], 20) #freeflow

def test_3link_queuespillback():

    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid1", 1, 1)
    W.addNode("mid2", 2, 2)
    W.addNode("dest", 3, 3)
    link1 = W.addLink("link1", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid1", "mid2", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.4)
    link3 = W.addLink("link3", "mid2", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    t1 = 400
    t2 = 800
    W.adddemand("orig", "dest", 0, t1, 0.8)
    W.adddemand("orig", "dest", t1, t2, 0.4)
    W.adddemand("orig", "dest", t2, 2000, 0.1)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()

    assert equal_tolerance(W.analyzer.trip_all, 600)
    assert equal_tolerance(W.analyzer.trip_completed, 585)
    assert equal_tolerance(W.analyzer.total_travel_time, 221050)
    assert equal_tolerance(W.analyzer.average_travel_time, 378)
    assert equal_tolerance(W.analyzer.average_delay, 228)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[1, 8], 0.8, rel_tol=0.2) #before congestion
    assert equal_tolerance(link1.k_mat[1, 8], 0.04, rel_tol=0.2) #before congestion
    assert equal_tolerance(link1.v_mat[1, 8], 20, rel_tol=0.2) #before congestion
    assert equal_tolerance(link1.q_mat[5, 8], 0.4, rel_tol=0.2) #during congestion
    assert equal_tolerance(link1.k_mat[5, 8], 0.12, rel_tol=0.2) #during congestion
    assert equal_tolerance(link1.v_mat[5, 8], 3.33, rel_tol=0.2) #during congestion
    assert equal_tolerance(link1.q_mat[10, 8], 0.08, rel_tol=0.2) #after congestion
    assert equal_tolerance(link1.k_mat[10, 8], 0.004, rel_tol=0.2) #after congestion
    assert equal_tolerance(link1.v_mat[10, 8], 20, rel_tol=0.2) #after congestion

    assert equal_tolerance(link2.q_mat[5, 8], 0.4, rel_tol=0.2) #during congestion
    assert equal_tolerance(link2.k_mat[5, 8], 0.12, rel_tol=0.2) #during congestion
    assert equal_tolerance(link2.v_mat[5, 8], 3.33, rel_tol=0.2) #during congestion
    assert equal_tolerance(link2.q_mat[13, 8], 0.08, rel_tol=0.2) #after congestion
    assert equal_tolerance(link2.k_mat[13, 8], 0.004, rel_tol=0.2) #after congestion 
    assert equal_tolerance(link2.v_mat[13, 8], 20, rel_tol=0.2) #after congestion

    assert equal_tolerance(link3.q_mat[5, 8], 0.4, rel_tol=0.2) #during congestion
    assert equal_tolerance(link3.k_mat[5, 8], 0.02, rel_tol=0.2) #during congestion
    assert equal_tolerance(link3.v_mat[5, 8], 20, rel_tol=0.2) #during congestion
    assert equal_tolerance(link3.q_mat[15, 8], 0.08, rel_tol=0.2) #after congestion
    assert equal_tolerance(link3.k_mat[15, 8], 0.004, rel_tol=0.2) #after congestion 
    assert equal_tolerance(link3.v_mat[15, 8], 20, rel_tol=0.2) #after congestion

def test_2link_signal():
    W = World(
        name="",
        deltan=5, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1, signal=[60,60])
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 300, 800, 0.4)
    W.adddemand("orig", "dest", 0, 2000, 0.2)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 600)
    assert equal_tolerance(W.analyzer.trip_completed, 570)
    assert equal_tolerance(W.analyzer.total_travel_time, 107100)
    assert equal_tolerance(W.analyzer.average_travel_time, 188)
    assert equal_tolerance(W.analyzer.average_delay, 88)

    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[:,8].mean() , 0.29947916666666663)
    assert equal_tolerance(link1.k_mat[:,8].mean() , 0.054958767361111105)
    assert equal_tolerance(link1.v_mat[:,8].mean() , 12.86046633191759)
    assert equal_tolerance(link1.q_mat[:,2].mean() , 0.3020833333333333)
    assert equal_tolerance(link1.k_mat[:,2].mean() , 0.026356336805555557)
    assert equal_tolerance(link1.v_mat[:,2].mean() , 17.49905597926791)
    assert equal_tolerance(link1.q_mat[8,:].mean() , 0.309375)
    assert equal_tolerance(link1.k_mat[8,:].mean() , 0.06661458333333334)
    assert equal_tolerance(link1.v_mat[8,:].mean() , 9.633282431298722)
    assert equal_tolerance(link1.q_mat[12,:].mean(),  0.19687499999999997)
    assert equal_tolerance(link1.k_mat[12,:].mean(),  0.011875)
    assert equal_tolerance(link1.v_mat[12,:].mean(),  18.607142857142858)
    assert equal_tolerance(link2.q_mat[:,8].mean() , 0.29427083333333337)
    assert equal_tolerance(link2.k_mat[:,8].mean() , 0.014713541666666666)
    assert equal_tolerance(link2.v_mat[:,8].mean() , 20.0)


def test_2link_signal_deltan1():
    W = World(
        name="",
        deltan=1, 
        tmax=2000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid", 1, 1, signal=[60,60])
    W.addNode("dest", 2, 2)
    link1 = W.addLink("link1", "orig", "mid", length=1000, free_flow_speed=20, jam_density=0.2)
    link2 = W.addLink("link2", "mid", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 300, 800, 0.4)
    W.adddemand("orig", "dest", 0, 2000, 0.2)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    W.analyzer.basic_analysis()
    assert equal_tolerance(W.analyzer.trip_all, 600)
    assert equal_tolerance(W.analyzer.trip_completed, 568)
    assert equal_tolerance(W.analyzer.total_travel_time, 112348)
    assert equal_tolerance(W.analyzer.average_travel_time, 197)
    assert equal_tolerance(W.analyzer.average_delay, 97)
    
    W.analyzer.compute_edie_state()
    assert equal_tolerance(link1.q_mat[:,8].mean() , 0.2994791666666667)
    assert equal_tolerance(link1.k_mat[:,8].mean() , 0.06033072916666667)
    assert equal_tolerance(link1.v_mat[:,8].mean() , 11.848820125791734)
    assert equal_tolerance(link1.q_mat[:,2].mean() , 0.3026041666666667)
    assert equal_tolerance(link1.k_mat[:,2].mean() , 0.02812717013888889)
    assert equal_tolerance(link1.v_mat[:,2].mean() , 16.587738607378817)
    assert equal_tolerance(link1.q_mat[8,:].mean() , 0.3302083333333333)
    assert equal_tolerance(link1.k_mat[8,:].mean() , 0.07739166666666668)
    assert equal_tolerance(link1.v_mat[8,:].mean() , 7.4833973250262265)
    assert equal_tolerance(link1.q_mat[12,:].mean(),  0.2)
    assert equal_tolerance(link1.k_mat[12,:].mean(),  0.013500000000000002)
    assert equal_tolerance(link1.v_mat[12,:].mean(),  18.444444444444446)
    assert equal_tolerance(link2.q_mat[:,8].mean() , 0.29427083333333337)
    assert equal_tolerance(link2.k_mat[:,8].mean() , 0.014713541666666666)
    assert equal_tolerance(link2.v_mat[:,8].mean() , 20.0)

def test_KW_theory_cumulative_curves_and_travel_time():
    W = World(
        name="",
        deltan=5, 
        tmax=3000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=0
    )

    W.addNode("orig", 0, 0)
    W.addNode("bottleneck", 1, 1)
    W.addNode("dest", 1, 1)
    link1 = W.addLink("link1", "orig", "bottleneck", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.5)
    link2 = W.addLink("link2", "bottleneck", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
    W.adddemand("orig", "dest", 0, 500, 0.2)
    W.adddemand("orig", "dest", 500, 1000, 0.5)
    W.adddemand("orig", "dest", 1000, 1500, 0.6)
    W.adddemand("orig", "dest", 1500, 2000, 0.2)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    # W.analyzer.cumulative_curves(link1)

    assert equal_tolerance(link1.arrival_count(700) - link1.departure_count(700), 0.5/link1.u*link1.length)    #free-flow
    assert equal_tolerance(link1.actual_travel_time(700), link1.length/link1.u)    #free-flow
    assert link1.arrival_count(1500) - link1.departure_count(1500) > link1.k_star*link1.length    #congested
    assert link1.actual_travel_time(1500) > link1.length/link1.u    #congested
    assert equal_tolerance(link1.arrival_count(1500), 650)
    assert equal_tolerance(link1.departure_count(1500), 575)
    assert equal_tolerance(link1.arrival_count(1500), link1.departure_count(1500+link1.actual_travel_time(1500))) #congested
    for t in range(0, W.TMAX, 10):
        assert equal_tolerance(link1.arrival_count(t), link1.departure_count(t+link1.actual_travel_time(t)))

def rigorous_verification_of_KW_theory_cumulative_curves_and_travel_time(deltan):
    """
    This function is the most rigorous verification of UXsim, checking that UXsim correctly solves KW theory.
    Specifically, it checks the relation between the cumulative curves and the travel time of vehicles.
    """

    W = World(
        name="",
        deltan=deltan, 
        tmax=3000, 
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("bottleneck", 1, 1)
    W.addNode("dest", 1, 1)
    link1 = W.addLink("link1", "orig", "bottleneck", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.6)
    link2 = W.addLink("link2", "bottleneck", "dest", length=1000, free_flow_speed=20, jam_density=0.2)

    dt = 50
    for t in range(0, 2000, dt):
        W.adddemand("orig", "dest", t, t+dt, random.uniform(0,1.0))

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    #W.analyzer.cumulative_curves(link1)

    for t in range(0, W.TMAX, 10):
        assert equal_tolerance(link1.arrival_count(t), link1.departure_count(t+link1.actual_travel_time(t)), abs_tol=W.DELTAN)

    for vehid in W.VEHICLES:
        depart_t = W.VEHICLES[vehid].log_t_link[0][0]
        depart_timestep = int(depart_t/W.DELTAT)
        traveltime_recorded_by_vehicle = W.VEHICLES[vehid].travel_time
        traveltime_from_log_state = (W.VEHICLES[vehid].log_state.count("run") + W.VEHICLES[vehid].log_state.count("wait") )*W.DELTAT
        traveltime_from_log_t = W.VEHICLES[vehid].log_t[-1] - W.VEHICLES[vehid].log_t[depart_timestep] + W.DELTAT - W.DELTAT

        link_enter_t = W.VEHICLES[vehid].log_t_link[1][0]
        link_enter_timestep = int(link_enter_t/W.DELTAT)
        traveltime_from_cumulative_curves_onlink = int(link1.actual_travel_time(link_enter_t) + link2.actual_travel_time(link1.actual_travel_time(link_enter_t)))
        traveltime_from_log_state_onlink = (W.VEHICLES[vehid].log_state.count("run"))*W.DELTAT
        traveltime_from_log_t_onlink = W.VEHICLES[vehid].log_t[-1] - W.VEHICLES[vehid].log_t[link_enter_timestep] + W.DELTAT

        #trip travel time including waiting at the vertical queue
        assert traveltime_recorded_by_vehicle >= traveltime_from_log_state_onlink
        assert traveltime_recorded_by_vehicle == traveltime_from_log_state
        assert traveltime_recorded_by_vehicle == traveltime_from_log_t

        #within link travel time
        assert traveltime_from_log_state_onlink >= link1.length/link1.u
        assert traveltime_from_log_state_onlink == traveltime_from_log_t_onlink
        assert equal_tolerance(traveltime_from_log_state_onlink, traveltime_from_cumulative_curves_onlink, abs_tol=W.DELTAT)  #cumulative curve is approximation

@pytest.mark.flaky(reruns=2)
def test_KW_theory_cumulative_curves_and_travel_time():
    rigorous_verification_of_KW_theory_cumulative_curves_and_travel_time(5)

@pytest.mark.flaky(reruns=2)
def test_KW_theory_cumulative_curves_and_travel_time_deltan1():
    rigorous_verification_of_KW_theory_cumulative_curves_and_travel_time(1)

def test_iterative_exec_rigorous():
    for _ in range(20):
        W = World(
            name="",
            deltan=10,
            tmax=100,
            duo_update_time=10,
            print_mode=0, save_mode=0, show_mode=0,
            random_seed=42,
        )

        W.addNode("orig", 0, 0)
        W.addNode("dest", 2, 1)
        W.addLink("link1", "orig", "dest", length=10000, free_flow_speed=20, number_of_lanes=1)
        W.addVehicle("orig", "dest", 0)
        W.addVehicle("orig", "dest", 50)
        W.addVehicle("orig", "dest", 100)

        #print("W.T, start_t, end_t")
        print("W.T", "\t", "W.TIME", "duration_t", "ongoing")
        
        if random.random() < 0.5:
            maxt = 30
        else:
            maxt = 200

        while W.check_simulation_ongoing():
            duration_t = random.randint(0, maxt)
            if hasattr(W, "T"):
                print(W.T, "\t", W.TIME, "\t", duration_t, "\t", W.check_simulation_ongoing())
            else:
                print(0, "\t", 0, "\t", duration_t, "\t", W.check_simulation_ongoing())
            W.exec_simulation(duration_t=duration_t)
        else:
            print(W.T, "\t","\t",  "\t", W.check_simulation_ongoing())


        if  hasattr(W.analyzer, "total_distance_traveled"):
            print("SUCCESS TO TERMINATE")
            W.analyzer.print_simple_stats(force_print=True)
        else:
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            assert False
        print_columns(["log_t"]+W.VEHICLES["0"].log_t, ["log_x"]+W.VEHICLES["0"].log_x, ["log_v"]+W.VEHICLES["0"].log_v)
        print("final_x", W.VEHICLES["0"].x)
        assert W.VEHICLES["0"].log_t[-1] == 90
        assert W.VEHICLES["0"].log_x[-1] == 1600
        assert W.VEHICLES["0"].x == 1800
        """
        This simulation runs until 100 sec with deltat=10 sec.
        Thus, the final time step start from 90 sec and simulate the duration of [90 sec, 100 sec).
        The last recorded vehicle position in `log_x` is 1600 m that is the position on 90 sec ((90 sec - 10 sec) * 20 m/sec) where the departure time was 10 sec.
        The last position `x` is 1800 m that is the "to-be" position on time 100 sec.
        """

        print()


def test_iterative_exec_rigorous_random_size_old():
    for _ in range(100):
        deltan = random.randint(1,10)
        tmax = random.randint(50,150)
        link_u = random.randint(10,30)
        W = World(
            name="",
            deltan=deltan,
            tmax=tmax,
            duo_update_time=10,
            print_mode=1, save_mode=0, show_mode=0,
            random_seed=42,
        )
        if not tmax//W.DELTAT-2 > 1:
            continue

        W.addNode("orig", 0, 0)
        W.addNode("dest", 2, 1)
        W.addLink("link1", "orig", "dest", length=10000, free_flow_speed=link_u, number_of_lanes=1)
        W.addVehicle("orig", "dest", 0)
        W.addVehicle("orig", "dest", 50)
        W.addVehicle("orig", "dest", 100)

        #print("W.T, start_t, end_t")
        # print("W.T", "\t", "W.TIME", "duration_t", "ongoing")
        
        if random.random() < 0.5:
            maxt = 30
        else:
            maxt = 200

        while W.check_simulation_ongoing():
            duration_t = random.randint(0, maxt)
            # if hasattr(W, "T"):
            #     print(W.T, "\t", W.TIME, "\t", duration_t, "\t", W.check_simulation_ongoing())
            # else:
            #     print(0, "\t", 0, "\t", duration_t, "\t", W.check_simulation_ongoing())
            W.exec_simulation(duration_t=duration_t)
        # else:
        #     print(W.T, "\t","\t",  "\t", W.check_simulation_ongoing())

        if  hasattr(W.analyzer, "total_distance_traveled"):
            print("SUCCESS TO TERMINATE")
            W.analyzer.print_simple_stats(force_print=True)
        else:
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            assert False
        print_columns(["log_t"]+W.VEHICLES["0"].log_t, ["log_x"]+W.VEHICLES["0"].log_x, ["log_v"]+W.VEHICLES["0"].log_v)
        print("final_x", W.VEHICLES["0"].x)
        assert W.VEHICLES["0"].log_t[-1] == (tmax//W.DELTAT-1)*W.DELTAT
        assert W.VEHICLES["0"].log_x[-1] == (tmax//W.DELTAT-2)*W.DELTAT*link_u
        assert W.VEHICLES["0"].x == (tmax//W.DELTAT-1)*W.DELTAT*link_u

        print()


def test_iterative_exec_rigorous_random_size_duration_t2():
    for _ in range(100):
        deltan = random.randint(1,10)
        tmax = random.randint(50,150)
        link_u = random.randint(10,30)
        W = World(
            name="",
            deltan=deltan,
            tmax=tmax,
            duo_update_time=10,
            print_mode=1, save_mode=0, show_mode=0,
            random_seed=42,
        )
        if not tmax//W.DELTAT-2 > 1:
            continue

        W.addNode("orig", 0, 0)
        W.addNode("dest", 2, 1)
        W.addLink("link1", "orig", "dest", length=10000, free_flow_speed=link_u, number_of_lanes=1)
        W.addVehicle("orig", "dest", 0)
        W.addVehicle("orig", "dest", 50)
        W.addVehicle("orig", "dest", 100)

        #print("W.T, start_t, end_t")
        # print("W.T", "\t", "W.TIME", "duration_t2", "ongoing")
        
        if random.random() < 0.5:
            maxt = 30
        else:
            maxt = 200

        while W.check_simulation_ongoing():
            duration_t2 = random.randint(deltan, maxt)
            # if hasattr(W, "T"):
            #     print(W.T, "\t", W.TIME, "\t", duration_t2, "\t", W.check_simulation_ongoing())
            # else:
            #     print(0, "\t", 0, "\t", duration_t2, "\t", W.check_simulation_ongoing())
            W.exec_simulation(duration_t2=duration_t2)
        # else:
        #     print(W.T, "\t","\t",  "\t", W.check_simulation_ongoing())

        if  hasattr(W.analyzer, "total_distance_traveled"):
            print("SUCCESS TO TERMINATE")
            W.analyzer.print_simple_stats(force_print=True)
        else:
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("FAILED TO TERMINATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            assert False
        print_columns(["log_t"]+W.VEHICLES["0"].log_t, ["log_x"]+W.VEHICLES["0"].log_x, ["log_v"]+W.VEHICLES["0"].log_v)
        print("final_x", W.VEHICLES["0"].x)
        assert W.VEHICLES["0"].log_t[-1] == (tmax//W.DELTAT-1)*W.DELTAT
        assert W.VEHICLES["0"].log_x[-1] == (tmax//W.DELTAT-2)*W.DELTAT*link_u
        assert W.VEHICLES["0"].x == (tmax//W.DELTAT-1)*W.DELTAT*link_u

        print()

def test_vehicle_speed_at_node_transfer():
    for i in range(10):    
        if i == 0:
            deltan = 5
            free_flow_speed = 10
            length1 = 1000
            length2 = random.randint(500,1000)
            length3 = 1000
            lanes1 = 1
            lanes2 = 1
            lanes3 = 1
        else:
            deltan = random.randint(1,10)
            free_flow_speed = random.randint(5,25)
            length1 = random.randint(500,1000)
            length2 = random.randint(500,1000)
            length3 = random.randint(500,1000)
            lanes1 = random.randint(1,5)
            lanes2 = random.randint(1,5)
            lanes3 = random.randint(1,5)
            
        W = World(
            name="",
            deltan=deltan,
            tmax=3600,
            print_mode=1, save_mode=0, show_mode=0,
            random_seed=0
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", 1, 1)
        W.addNode("dest", 2, 1)
        W.addLink(f"link1-{length1}", "orig", "mid1", length=length1, free_flow_speed=free_flow_speed, number_of_lanes=lanes1)
        W.addLink(f"link2-{length2}", "mid1", "mid2", length=length2, free_flow_speed=free_flow_speed, number_of_lanes=lanes2)
        W.addLink(f"link3-{length3}", "mid2", "dest", length=length3, free_flow_speed=free_flow_speed, number_of_lanes=lanes3)
        W.adddemand("orig", "dest", 0, 1000, 0.1)
        W.exec_simulation()

        W.analyzer.print_simple_stats()
        df = W.analyzer.vehicles_to_pandas()
        df = df[(df["name"]=="0") & (df["x"]!=-1)]
        print(df)
        vs = df["v"]
        for v in list(vs):
            assert equal_tolerance(v, free_flow_speed, rel_tol=0, abs_tol=0.01)
                    
